"""Enum class containing the valid values for the Spatial parameter category
of the external forcings.

for more details check D-Flow FM User Manual 1D2D, Chapter D.3.1, Table D.2
https://content.oss.deltares.nl/delft3d/D-Flow_FM_User_Manual_1D2D.pdf
"""

from enum import IntEnum
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import yaml
from pydantic import Field, field_validator, model_validator
from strenum import StrEnum

from hydrolib.core.base.models import (
    BaseModel,
    DiskOnlyFileModel,
    ModelSaveSettings,
    ParsableFileModel,
    SerializerConfig,
)
from hydrolib.core.base.utils import resolve_file_model
from hydrolib.core.dflowfm.common.models import Operand
from hydrolib.core.dflowfm.extold.parser import Parser
from hydrolib.core.dflowfm.extold.serializer import Serializer
from hydrolib.core.dflowfm.ini.util import enum_value_parser
from hydrolib.core.dflowfm.polyfile.models import PolyFile
from hydrolib.core.dflowfm.tim.models import TimModel

EXT_OLD_MODULE_PATH = Path(__file__).parent / "data"

with (EXT_OLD_MODULE_PATH / "old-external-forcing-data.yaml").open("r") as fh:
    QUANTITIES_DATA = yaml.safe_load(fh)

with (EXT_OLD_MODULE_PATH / "header.txt").open("r") as f:
    HEADER = f.read()

INITIALTRACER = "initialtracer"

FILETYPE_FILEMODEL_MAPPING = {
    1: TimModel,
    2: TimModel,
    3: DiskOnlyFileModel,
    4: DiskOnlyFileModel,
    5: DiskOnlyFileModel,
    6: DiskOnlyFileModel,
    7: DiskOnlyFileModel,
    8: DiskOnlyFileModel,
    9: PolyFile,
    10: PolyFile,
    11: DiskOnlyFileModel,
    12: DiskOnlyFileModel,
}

BOUNDARY_CONDITION_QUANTITIES_VALID_PREFIXES = tuple(
    QUANTITIES_DATA["BoundaryCondition"]["prefixes"]
)


class _ExtOldBoundaryQuantity(StrEnum):

    @classmethod
    def _missing_(cls, value):
        """Custom implementation for handling missing values.

        the method parses any missing values and only allows the ones that start with "initialtracer".
        """
        # Allow strings starting with "tracer"
        if isinstance(value, str) and value.startswith(
            BOUNDARY_CONDITION_QUANTITIES_VALID_PREFIXES
        ):
            new_member = str.__new__(cls, value)
            new_member._value_ = value
            return new_member
        else:
            raise ValueError(
                f"{value} is not a valid {cls.__name__} possible quantities are {', '.join(cls.__members__)}, "
                f"and quantities that start with 'tracer'"
            )


ExtOldBoundaryQuantity = StrEnum(
    "ExtOldBoundaryQuantity",
    QUANTITIES_DATA["BoundaryCondition"]["quantity_names"],
    type=_ExtOldBoundaryQuantity,
)
ExtOldParametersQuantity = StrEnum(
    "ExtOldParametersQuantity", QUANTITIES_DATA["Parameter"]["quantity_names"]
)
PARAMETER_QUANTITIES_VALID_PREFIXES = tuple(QUANTITIES_DATA["Parameter"]["prefixes"])
ExtOldMeteoQuantity = StrEnum(
    "ExtOldMeteoQuantity",
    QUANTITIES_DATA["Meteo"]["quantity_names"],
)
ExtOldLateralQuantity = StrEnum("ExtOldLateralQuantity", QUANTITIES_DATA["Lateral"])


INITIAL_CONDITION_QUANTITIES_VALID_PREFIXES = tuple(
    QUANTITIES_DATA["InitialConditions"]["prefixes"]
)


class _ExtOldInitialConditionQuantity(StrEnum):
    """
    If there is a missing quantity that is mentioned in the "Accepted quantity names" section of the user manual
    [Sec.C.5.3](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.5.3).
    and [Sec.D.3](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.D.3).
    please open and issue in github.
    """

    @classmethod
    def _missing_(cls, value):
        """Custom implementation for handling missing values.

        the method parses any missing values and only allows the ones that start with "initialtracer".
        """
        # Allow strings starting with "tracer"
        if isinstance(value, str) and value.startswith(
            INITIAL_CONDITION_QUANTITIES_VALID_PREFIXES
        ):
            new_member = str.__new__(cls, value)
            new_member._value_ = value
            return new_member
        else:
            raise ValueError(
                f"{value} is not a valid {cls.__name__} possible quantities are {', '.join(cls.__members__)}, "
                f"and quantities that start with 'tracer'"
            )


ExtOldInitialConditionQuantity = StrEnum(
    "ExtOldInitialConditionQuantity",
    QUANTITIES_DATA["InitialConditions"]["quantity_names"],
    type=_ExtOldInitialConditionQuantity,
)

ExtOldSourcesSinks = StrEnum("ExtOldSourcesSinks", QUANTITIES_DATA["SourceSink"])

ALL_QUANTITIES = (
    QUANTITIES_DATA["BoundaryCondition"]["quantity_names"]
    | QUANTITIES_DATA["Meteo"]["quantity_names"]
    | QUANTITIES_DATA["Parameter"]["quantity_names"]
    | QUANTITIES_DATA["InitialConditions"]["quantity_names"]
    | QUANTITIES_DATA["SourceSink"]
    | QUANTITIES_DATA["Structure"]
    | QUANTITIES_DATA["Misellaneous"]
    | QUANTITIES_DATA["Lateral"]
)

ALL_PREFIXES = (
    BOUNDARY_CONDITION_QUANTITIES_VALID_PREFIXES
    + INITIAL_CONDITION_QUANTITIES_VALID_PREFIXES
    + PARAMETER_QUANTITIES_VALID_PREFIXES
)

ExtOldQuantity = StrEnum("ExtOldQuantity", ALL_QUANTITIES)


ExtOldFileType = IntEnum("ExtOldFileType", QUANTITIES_DATA["FileType"])
ExtOldMethod = IntEnum("ExtOldMethod", QUANTITIES_DATA["OldMethods"])
ExtOldExtrapolationMethod = IntEnum(
    "ExtOldExtrapolationMethod", QUANTITIES_DATA["ExtrapolationMethod"]
)


class ExtOldForcing(BaseModel):
    """Class holding the external forcing values.

    This class is used to represent the external forcing values in the D-Flow FM model.

    Attributes:
        quantity (Union[ExtOldQuantity, str]):
            The name of the quantity.
        filename (Union[PolyFile, TimModel, DiskOnlyFileModel]):
            The file associated with this forcing.
        varname (Optional[str]):
            The variable name used in `filename` associated with this forcing; some input files may contain multiple variables.
        sourcemask (DiskOnlyFileModel):
            The file containing a mask.
        filetype (ExtOldFileType):
            Indication of the file type.
            Options:
                1. Time series
                2. Time series magnitude and direction
                3. Spatially varying weather
                4. ArcInfo
                5. Spiderweb data (cyclones)
                6. Curvilinear data
                7. Samples (C.3)
                8. Triangulation magnitude and direction
                9. Polyline (<*.pli>-file, C.2)
                11. NetCDF grid data (e.g. meteo fields)
                14. NetCDF wave data
        method (ExtOldMethod):
            The method of interpolation.
            Options:
                1. Pass through (no interpolation)
                2. Interpolate time and space
                3. Interpolate time and space, save weights
                4. Interpolate space
                5. Interpolate time
                6. Averaging space
                7. Interpolate/Extrapolate time
        extrapolation_method (Optional[ExtOldExtrapolationMethod]):
            The extrapolation method.
                Options:
                    0. No spatial extrapolation.
                    1. Do spatial extrapolation outside of source data bounding box.
        maxsearchradius (Optional[float]):
            Search radius for model grid points that lie outside of the source data bounding box.
        operand (Operand):
            The operand to use for adding the provided values.

            Options:
                'O' Existing values are overwritten with the provided values.
                'A' Provided values are used where existing values are missing.
                '+' Existing values are summed with the provided values.
                '*' Existing values are multiplied with the provided values.
                'X' The maximum values of the existing values and provided values are used.
                'N' The minimum values of the existing values and provided values are used.
        value (Optional[float]):
            Custom coefficients for transformation.
        factor (Optional[float]):
            The conversion factor.
        ifrctyp (Optional[float]):
            The friction type.
        averagingtype (Optional[float]):
            The averaging type.
        relativesearchcellsize (Optional[float]):
            The relative search cell size for samples inside a cell.
        extrapoltol (Optional[float]):
            The extrapolation tolerance.
        percentileminmax (Optional[float]):
            Changes the min/max operator to an average of the highest/lowest data points.
            The value sets the percentage of the total set that is to be included.
        area (Optional[float]):
            The area for sources and sinks.
        nummin (Optional[int]):
            The minimum required number of source data points in each target cell.
    """

    quantity: Union[ExtOldQuantity, str] = Field(alias="QUANTITY")
    filename: Union[PolyFile, TimModel, DiskOnlyFileModel] = Field(
        None, alias="FILENAME"
    )
    varname: Optional[str] = Field(None, alias="VARNAME")
    sourcemask: DiskOnlyFileModel = Field(
        default_factory=lambda: DiskOnlyFileModel(None), alias="SOURCEMASK"
    )
    filetype: ExtOldFileType = Field(alias="FILETYPE")
    method: ExtOldMethod = Field(alias="METHOD")
    extrapolation_method: Optional[ExtOldExtrapolationMethod] = Field(
        None, alias="EXTRAPOLATION_METHOD"
    )

    maxsearchradius: Optional[float] = Field(None, alias="MAXSEARCHRADIUS")
    operand: Operand = Field(alias="OPERAND")
    value: Optional[float] = Field(None, alias="VALUE")
    factor: Optional[float] = Field(None, alias="FACTOR")
    ifrctyp: Optional[float] = Field(None, alias="IFRCTYP")
    averagingtype: Optional[float] = Field(None, alias="AVERAGINGTYPE")

    relativesearchcellsize: Optional[float] = Field(
        None, alias="RELATIVESEARCHCELLSIZE"
    )
    extrapoltol: Optional[float] = Field(None, alias="EXTRAPOLTOL")
    percentileminmax: Optional[float] = Field(None, alias="PERCENTILEMINMAX")
    area: Optional[float] = Field(None, alias="AREA")
    nummin: Optional[int] = Field(None, alias="NUMMIN")

    tracerfallvelocity: Optional[float] = Field(None, alias="TRACERFALLVELOCITY")
    tracerdecaytime: Optional[float] = Field(None, alias="TRACERDECAYTIME")

    def is_intermediate_link(self) -> bool:
        return True

    @model_validator(mode="before")
    @classmethod
    def handle_case_insensitive_tracer_fields(cls, values):
        """Handle case-insensitive matching for tracer fields."""
        values_copy = dict(values)

        # Define the field names and their aliases
        tracer_fields = ["tracerfallvelocity", "tracerdecaytime"]

        for field_i in values.keys():
            if field_i.lower() in tracer_fields:
                # if the field is already lowercase no need to change it
                if field_i != field_i.lower():
                    # If the key is not in the expected lowercase format, add it with the correct format
                    values_copy[field_i.lower()] = values_copy[field_i]
                    # Remove the original key to avoid "extra fields not permitted" error
                    values_copy.pop(field_i)

        return values_copy

    @classmethod
    def validate_quantity_prefix(
        cls, lower_value: str, value_str: str
    ) -> Optional[str]:
        """Checks if the provided quantity string starts with any known valid prefix.

        If the quantity matches a prefix, ensures it is followed by a name.
        Returns the full quantity string if valid, otherwise None.

        Args:
            lower_value (str): The quantity string in lowercase.
            value_str (str): The original quantity string.

        Raises:
            ValueError: If the quantity is only the prefix without a name.
        """
        value = None
        for prefix in ALL_PREFIXES:
            if lower_value.startswith(prefix):
                n = len(prefix)
                if n == len(value_str):
                    raise ValueError(
                        f"QUANTITY '{value_str}' should be appended with a valid name."
                    )
                value = prefix + value_str[n:]
                break

        return value

    @field_validator("quantity", mode="before")
    @classmethod
    def validate_quantity(cls, value) -> Any:
        if not isinstance(value, ExtOldQuantity):
            found = False
            value_str = str(value)
            lower_value = value_str.lower()

            quantity = cls.validate_quantity_prefix(lower_value, value_str)
            if quantity is not None:
                value = quantity
                found = True
            elif lower_value in list(ExtOldQuantity):
                value = ExtOldQuantity(lower_value)
                found = True
            elif value_str in list(ExtOldQuantity):
                value = ExtOldQuantity(value_str)
                found = True

            if not found:
                supported_value_str = ", ".join(([x.value for x in ExtOldQuantity]))
                raise ValueError(
                    f"QUANTITY '{value_str}' not supported. Supported values: {supported_value_str}"
                )
        return value

    @field_validator("operand", mode="before")
    @classmethod
    def validate_operand(cls, value):
        return enum_value_parser(value, Operand)

    @model_validator(mode="after")
    def validate_varname(self):
        if self.varname and self.filetype != ExtOldFileType.NetCDFGridData:
            raise ValueError(
                "VARNAME only allowed when FILETYPE is 11 (NetCDFGridData)"
            )
        return self

    @field_validator("extrapolation_method")
    @classmethod
    def validate_extrapolation_method(cls, v, info):
        method = info.data.get("method")
        valid_extrapolation_method = (
            ExtOldExtrapolationMethod.SpatialExtrapolationOutsideOfSourceDataBoundingBox
        )
        available_extrapolation_methods = [
            ExtOldMethod.InterpolateTimeAndSpaceSaveWeights,
            ExtOldMethod.Obsolete,
        ]
        if (
            v == valid_extrapolation_method
            and method not in available_extrapolation_methods
        ):
            raise ValueError(
                f"EXTRAPOLATION_METHOD only allowed to be {valid_extrapolation_method} when METHOD is "
                f"{available_extrapolation_methods[0]} or {available_extrapolation_methods[1]}"
            )
        return v

    @model_validator(mode="after")
    def validate_factor(self):
        quantity = self.quantity
        if self.factor is not None and not str(quantity).startswith(INITIALTRACER):
            raise ValueError(
                f"FACTOR only allowed when QUANTITY starts with {INITIALTRACER}"
            )
        return self

    @model_validator(mode="after")
    def validate_ifrctyp(self):
        if (
            self.ifrctyp is not None
            and self.quantity != ExtOldQuantity.FrictionCoefficient
        ):
            raise ValueError(
                f"IFRCTYP only allowed when QUANTITY is {ExtOldQuantity.FrictionCoefficient}"
            )
        return self

    @model_validator(mode="after")
    def validate_averagingtype(self):
        if (
            self.averagingtype is not None
            and self.method != ExtOldMethod.AveragingSpace
        ):
            raise ValueError(
                f"AVERAGINGTYPE only allowed when METHOD is {ExtOldMethod.AveragingSpace}"
            )
        return self

    @model_validator(mode="after")
    def validate_relativesearchcellsize(self):
        if (
            self.relativesearchcellsize is not None
            and self.method != ExtOldMethod.AveragingSpace
        ):
            raise ValueError(
                f"RELATIVESEARCHCELLSIZE only allowed when METHOD is {ExtOldMethod.AveragingSpace}"
            )
        return self

    @model_validator(mode="after")
    def validate_extrapoltol(self):
        if self.extrapoltol is not None and self.method != ExtOldMethod.InterpolateTime:
            raise ValueError("EXTRAPOLTOL only allowed when METHOD is 5")
        return self

    @model_validator(mode="after")
    def validate_percentileminmax(self):
        if (
            self.percentileminmax is not None
            and self.method != ExtOldMethod.AveragingSpace
        ):
            raise ValueError(
                f"PERCENTILEMINMAX only allowed when METHOD is {ExtOldMethod.AveragingSpace}"
            )
        return self

    @model_validator(mode="after")
    def validate_area(self):
        if (
            self.area is not None
            and self.quantity != ExtOldQuantity.DischargeSalinityTemperatureSorSin
        ):
            raise ValueError(
                f"AREA only allowed when QUANTITY is {ExtOldQuantity.DischargeSalinityTemperatureSorSin}"
            )
        return self

    @model_validator(mode="after")
    def validate_nummin(self):
        if self.nummin is not None and self.method != ExtOldMethod.AveragingSpace:
            raise ValueError(
                f"NUMMIN only allowed when METHOD is {ExtOldMethod.AveragingSpace}"
            )
        return self

    @field_validator("maxsearchradius")
    def validate_maxsearchradius(cls, v, info):
        if v is not None:
            extrap = info.data.get("extrapolation_method")
            extrapolation_method_value = (
                ExtOldExtrapolationMethod.SpatialExtrapolationOutsideOfSourceDataBoundingBox
            )
            if extrap != extrapolation_method_value:
                raise ValueError(
                    f"MAXSEARCHRADIUS only allowed when EXTRAPOLATION_METHOD is {extrapolation_method_value}"
                )
        return v

    @field_validator("value")
    @classmethod
    def validate_value(cls, v, info):
        if v is not None:
            method = info.data.get("method")
            if method != ExtOldMethod.InterpolateSpace:
                raise ValueError(
                    f"VALUE only allowed when METHOD is {ExtOldMethod.InterpolateSpace} (InterpolateSpace)"
                )
        return v

    @model_validator(mode="before")
    @classmethod
    def choose_file_model(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Root-level validator to the right class for the filename parameter based on the filetype.

        The validator chooses the right class for the filename parameter based on the FileType_FileModel_mapping
        dictionary.

        FileType_FileModel_mapping = {
            1: TimModel,
            2: TimModel,
            3: DiskOnlyFileModel,
            4: DiskOnlyFileModel,
            5: DiskOnlyFileModel,
            6: DiskOnlyFileModel,
            7: DiskOnlyFileModel,
            8: DiskOnlyFileModel,
            9: PolyFile,
            10: PolyFile,
            11: DiskOnlyFileModel,
            12: DiskOnlyFileModel,
        }
        """
        # if the filetype and the filename are present in the values
        if any(par in values for par in ["filetype", "FILETYPE"]) and any(
            par in values for par in ["filename", "FILENAME"]
        ):
            file_type_var_name = "filetype" if "filetype" in values else "FILETYPE"
            filename_var_name = "filename" if "filename" in values else "FILENAME"
            file_type = values.get(file_type_var_name)
            raw_path = values.get(filename_var_name)

            if isinstance(raw_path, (Path, str)):
                model = FILETYPE_FILEMODEL_MAPPING.get(int(file_type))
                values[filename_var_name] = resolve_file_model(raw_path, model)

        return values

    @model_validator(mode="before")
    @classmethod
    def validate_sourcemask(cls, data: Any) -> Any:
        filetype = data.get("filetype")
        sourcemask = data.get("sourcemask")

        # Convert string to DiskOnlyFileModel if needed
        if isinstance(sourcemask, str):
            data["sourcemask"] = DiskOnlyFileModel(sourcemask)
            sourcemask = data["sourcemask"]

        if sourcemask and filetype not in [
            ExtOldFileType.ArcInfo,
            ExtOldFileType.CurvilinearData,
        ]:
            raise ValueError("SOURCEMASK only allowed when FILETYPE is 4 or 6")
        return data


class ExtOldModel(ParsableFileModel):
    """
    The overall external forcings model that contains the contents of one external forcings file (old format).

    This model is typically referenced under a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.external_forcing.extforcefile`.

    Attributes:
        comment (List[str]):
            The comments in the header of the external forcing file.
        forcing (List[ExtOldForcing]):
            The external forcing/QUANTITY blocks in the external forcing file.
    """

    comment: List[str] = Field(default=HEADER.splitlines()[1:])
    forcing: List[ExtOldForcing] = Field(default_factory=list)

    @classmethod
    def _ext(cls) -> str:
        return ".ext"

    @classmethod
    def _filename(cls) -> str:
        return "externalforcings"

    @classmethod
    def _get_serializer(
        cls,
    ) -> Callable[[Path, Dict, SerializerConfig, ModelSaveSettings], None]:
        return Serializer.serialize

    @classmethod
    def _get_parser(cls) -> Callable[[Path], Dict]:
        return Parser.parse

    @property
    def quantities(self) -> List[str]:
        """List all the quantities in the external forcings file."""
        return [forcing.quantity for forcing in self.forcing]



"""T3D file"""

import re
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple

from pydantic import Field, field_validator, model_validator
from strenum import StrEnum

from hydrolib.core.base.models import (
    BaseModel,
    ModelSaveSettings,
    ParsableFileModel,
    SerializerConfig,
)
from hydrolib.core.dflowfm.t3d.parser import T3DParser
from hydrolib.core.dflowfm.t3d.serializer import T3DSerializer


class LayerType(StrEnum):
    """
    Layer types in the t3d file.
    """

    sigma = "SIGMA"
    z = "Z"


# Group1 (float at the beginning)
# Group2 (time unit) seconds|minutes|hours|days
# Group3 (date) YYYY-MM-DD
TIME_PATTERN = re.compile(
    r"^"
    r"(\d+(?:\.\d+)?(?:[eE][+\-]?\d+)?)\s+"  # (1) sci_float = \d+(?:\.\d+)?(?:[eE][+\-]?\d+)?  (handles 1, 1.23, 1e9, 1.23e-10, etc.)
    r"(seconds|minutes|hours|days)\s+"  # (2) space + valid time unit
    r"since\s+"  # (3) space + 'since' + space
    r"(\d{4}-\d{2}-\d{2}"  # (4) date in YYYY-MM-DD e.g. 2006-01-01 (|<---the parentheses at the beginning)
    r"(?:[T ][0-9]{2}:[0-9]{2}:[0-9]{2})?"  # (5) optional time: preceded by 'T' or space, then hh:mm:ss 'T00:00:00' or ' 00:00:00'
    r"(?:\s*(?:Z|[+\-][0-9]{2}:[0-9]{2}))?)"  # (6)(7) optional whitespace + optional timezone: Z or ±hh:mm. (|<---the parentheses ends here)
    r"$"
)


class T3DTimeRecord(BaseModel):
    """Single tim record, representing a time and a list of data.

    Args:
        time (str):
            A time string of the format: '<float> <time unit> since <ISO date> [optional time] [optional timezone]'.
            Example: '15.5 seconds since 2023-02-01T13:45:59Z'
            - the time unit is [seconds|minutes|hours|days].
            - ISO date YYYY-MM-DD.
            - With optional time and optional timezone.
            - i.e, "15.5 seconds since 2023-02-01T13:45:59Z"

        data (List[float]):
            Record of the time record.

    Raises:
        ValueError: If the time string is not in the expected format.

    Examples:
        ```python
        >>> record = T3DTimeRecord(time="15.5 seconds since 2023-02-01T13:45:59Z", data=[1, 2, 3, 4, 5])
        >>> record
        T3DTimeRecord(time='15.5 seconds since 2023-02-01T13:45:59Z', data=[1.0, 2.0, 3.0, 4.0, 5.0])

        ```
    """

    time: str = Field(
        ...,
        description=(
            "A time string of the format: '<float> <time unit> since <ISO date> [optional time] [optional timezone]'. "
            "Example: '15.5 seconds since 2023-02-01T13:45:59Z'"
        ),
    )

    data: List[float] = Field(default_factory=list)

    @field_validator("time", mode="before", check_fields=True)
    def validate_time_format(cls, value: str) -> str:
        """Check if the time string is in the expected format.

        - YYYY-MM-DD
        - Optional time, e.g. T12:34:56 or  12:34:56
        - Optional timezone: Z or +hh:mm or -hh:mm
        """
        if not TIME_PATTERN.match(value):
            raise ValueError(
                f"Time string '{value}' is not in the expected format:\n"
                "    <float> <time unit> since <ISO date> [optional time] [optional timezone]"
            )
        return value

    @property
    def time_unit(self) -> str:
        """Return the unit of the time string.

        Returns:
            str: The unit of the time string. The unit can be [seconds|minutes|hours|days].
        """
        match = TIME_PATTERN.match(self.time)
        return match.group(2)

    @property
    def reference_date(self) -> str:
        """Return the date of the time string.

        Returns:
            str: The date of the time string in the format: YYYY-MM-DD.
            - ISO date YYYY-MM-DD.
            - With optional time and optional timezone.
            - i.e, "15.5 seconds since 2023-02-01T13:45:59Z"
        """
        match = TIME_PATTERN.match(self.time)
        return match.group(3)

    @property
    def time_offset(self) -> float:
        """Return the time offset of the time string."""
        match = TIME_PATTERN.match(self.time)
        return float(match.group(1))


class T3DModel(ParsableFileModel):
    r"""T3D file model.

    Args:
        comments (List[str]):
            A list with the header comment of the tim file.
        records (List[T3DTimeRecord]):
            List of time records.
        layers (List[float]):
            List of layers.
        vectormax (Optional[int]):
            The VECTORMAX value. Default is None.
        layer_type (LayerType):
            The layer type.
        quantities_names (Optional[List[str]]):
            List of names for the quantities in the timeseries.

    Examples:
        ```python
        >>> from hydrolib.core.dflowfm.t3d.models import T3DTimeRecord, T3DModel
        >>> layer_name = "SIGMA"
        >>> comments = ["comment1", "comment2"]
        >>> layers = [1, 2, 3, 4, 5]
        >>> record = [
        ...     T3DTimeRecord(time="0 seconds since 2006-01-01 00:00:00 +00:00", data=[5.0, 5.0, 10.0, 10.0]),
        ...     T3DTimeRecord(time="1e9 seconds since 2001-01-01 00:00:00 +00:00", data=[5.0, 5.0, 10.0, 10.0])
        ... ]
        >>> model = T3DModel(comments=comments, layer_type=layer_name, layers=layers, records=record)
        >>> print(model) # doctest: +SKIP
        T3DModel(
            filepath=None,
            serializer_config=SerializerConfig(float_format=''),
            comments=['comment1', 'comment2'],
            records=[
                T3DTimeRecord(time='0 seconds since 2006-01-01 00:00:00 +00:00', data=[5.0, 5.0, 10.0, 10.0]),
                T3DTimeRecord(time='1e9 seconds since 2001-01-01 00:00:00 +00:00', data=[5.0, 5.0, 10.0, 10.0])
            ],
            layers=[1.0, 2.0, 3.0, 4.0, 5.0], vectormax=None, layer_type='SIGMA', quantities_names=None
        )
        >>> print(model.size)
        (2, 4)
        >>> model.quantities_names = ["quantity-1", "quantity-2", "quantity-3"] # doctest: +SKIP
        Traceback (most recent call last):
            model.quantities_names = ["quantity-1", "quantity-2", "quantity-3"]
            ^^^^^^^^^^^^^^^^^^^^^^
          File "...\Lib\site-packages\pydantic\v1\main.py", line 397, in __setattr__
            raise ValidationError(errors, self.__class__)
        pydantic.v1.error_wrappers.ValidationError: 1 validation error for T3DModel
        __root__
          The number of quantities names must be equal to the number of values in the records. (type=value_error)
        >>> model.quantities_names = ["quantity-1", "quantity-2", "quantity-3", "quantity-4"]
        >>> print(model.quantities_names)
        ['quantity-1', 'quantity-2', 'quantity-3', 'quantity-4']

    ```
    """

    comments: List[str] = Field(default_factory=list)
    records: List[T3DTimeRecord] = Field(default_factory=list)
    layers: List[float] = Field(default_factory=list)
    vectormax: Optional[int] = Field(default=None, alias="VECTORMAX")
    layer_type: LayerType = Field(default=None, alias="LAYER_TYPE")
    quantities_names: Optional[List[str]] = Field(default=None)

    @field_validator("records", mode="after", check_fields=True)
    def validate_record_length(cls, value: List[T3DTimeRecord]):
        """Check if the records have the same length."""
        records = [v.data for v in value]
        if not all(len(sublist) == len(records[0]) for sublist in records):
            raise ValueError("All records must have the same length.")

        return value

    @model_validator(mode="after")
    def validate_quantities_names(cls, value: "T3DModel") -> "T3DModel":
        """
        Validate that the number of quantities names is equal to the number of values in the records.
        """
        record = value.records
        record_len = len(record[0].data)
        quantities_names = value.quantities_names
        if quantities_names is not None and len(quantities_names) != record_len:
            raise ValueError(
                "The number of quantities names must be equal to the number of values in the records."
            )
        return value

    @property
    def size(self) -> Tuple[int, int]:
        """Return the number ot time step * length of each record."""
        return len(self.records), len(self.records[0].data)

    def _ext(self) -> str:
        return ".t3d"

    def _filename(self) -> str:
        return "t3d"

    @classmethod
    def _get_parser(cls) -> Callable[[Path], Dict]:
        return T3DParser.parse

    @classmethod
    def _get_serializer(
        cls,
    ) -> Callable[[Path, Dict, SerializerConfig, ModelSaveSettings], None]:
        return T3DSerializer.serialize

    def as_dict(self) -> Dict[str, List[float]]:
        """Return the data from the records as a dictionary.

        Returns:
            data (Dict[str, List[float]]):
                A dictionary with the data from the records.

        Examples:
            ```python
            >>> from hydrolib.core.dflowfm.t3d.models import T3DTimeRecord, T3DModel
            >>> layer_name = "SIGMA"
            >>> comments = ["comment1", "comment2"]
            >>> layers = [0, 0.2, 0.6, 0.8, 1.0]
            >>> record = [
            ...     T3DTimeRecord(time="0 seconds since 2006-01-01 00:00:00 +00:00", data=[5.0, 5.0, 10.0, 10.0]),
            ...     T3DTimeRecord(time="1e9 seconds since 2001-01-01 00:00:00 +00:00", data=[5.0, 5.0, 10.0, 10.0])
            ... ]
            >>> model = T3DModel(comments=comments, layer_type=layer_name, layers=layers, records=record)
            >>> model.as_dict()
            {0.0: [5.0, 5.0, 10.0, 10.0], 1000000000.0: [5.0, 5.0, 10.0, 10.0]}

            ```
        """
        data = {}
        for record in self.records:
            match = TIME_PATTERN.match(record.time)
            offset_str, _, _ = match.groups()
            offset = float(offset_str)
            data[offset] = record.data

        return data

    def get_units(self):
        """Return the units for each quantity in the timeseries.

        Returns:
            List[str]: A list of units for each quantity in the timeseries.

        Examples:
            Create a `TimModel` object from a .tim file:
                ```python
                >>> from hydrolib.core.dflowfm.t3d.models import T3DTimeRecord, T3DModel
                >>> layer_name = "SIGMA"
                >>> comments = ["comment1", "comment2"]
                >>> layers = [0, 0.2, 0.6, 0.8, 1.0]
                >>> record = [
                ...     T3DTimeRecord(time="0 seconds since 2006-01-01 00:00:00 +00:00", data=[5.0, 5.0, 10, 10, 10]),
                ...     T3DTimeRecord(time="1e9 seconds since 2001-01-01 00:00:00 +00:00", data=[5.0, 5.0, 10, 10, 10])
                ... ]
                >>> model = T3DModel(comments=comments, layer_type=layer_name, layers=layers, records=record)
                >>> model.quantities_names = ["discharge", "waterlevel", "temperature", "salinity", "initialtracer"]
                >>> print(model.get_units())
                ['m3/s', 'm', 'degC', '1e-3', '-']

                ```
        """
        if self.quantities_names is None:
            return None
        return T3DModel._get_quantity_unit(self.quantities_names)



import logging
from abc import ABC
from enum import Enum
from inspect import isclass
from math import isnan
from typing import (
    Annotated,
    Any,
    Callable,
    List,
    Literal,
    Optional,
    Set,
    Type,
    Union,
    get_args,
    get_origin,
)

from pandas import DataFrame
from pydantic import (
    BeforeValidator,
    ConfigDict,
    Field,
    GetCoreSchemaHandler,
    field_validator,
    model_validator,
)
from pydantic.fields import FieldInfo
from pydantic_core import core_schema

from hydrolib.core import __version__ as version
from hydrolib.core.base.file_manager import FilePathStyleConverter
from hydrolib.core.base.models import (
    BaseModel,
    FileModel,
    ModelSaveSettings,
    ParsableFileModel,
)
from hydrolib.core.base.utils import FortranScientificNotationConverter
from hydrolib.core.dflowfm.ini.io_models import (
    CommentBlock,
    Document,
    Property,
    Section,
)
from hydrolib.core.dflowfm.ini.parser import Parser
from hydrolib.core.dflowfm.ini.serializer import (
    DataBlockINIBasedSerializerConfig,
    INISerializerConfig,
    write_ini,
)
from hydrolib.core.dflowfm.ini.util import UnknownKeywordErrorManager, make_list

logger = logging.getLogger(__name__)


class INIBasedModel(BaseModel, ABC):
    """INIBasedModel defines the base model for blocks/chapters
    inside an INIModel (*.ini file).

    - Abstract base class for representing INI-style configuration file blocks or chapters.
    - This class serves as the foundational model for handling blocks within INI configuration files.
    It supports creating instances from parsed INI sections, adding arbitrary fields, and ensuring
    well-defined serialization and deserialization behavior. Subclasses are expected to define
    specific behavior and headers for their respective INI blocks.

    Attributes:
        comments (Optional[Comments], optional):
            Comments for the model fields. Defaults to None.

    Raises:
        ValueError: If unknown fields are encountered during validation.

    See Also:
        BaseModel: The Pydantic base model extended by this class.
        INISerializerConfig: Provides configuration for INI serialization.


    Examples:
        Define a custom INI block subclass:
            ```python
            >>> from hydrolib.core.dflowfm.ini.models import INIBasedModel
            >>> class MyModel(INIBasedModel):
            ...     _header = "MyHeader"
            ...     field_a: str = "default_value"

            ```

        Parse an INI section:
            ```python
            >>> from hydrolib.core.dflowfm.ini.io_models import Section
            >>> section = Section(header="MyHeader", content=[{"key": "field_a", "value": "value"}])
            >>> model = MyModel.model_validate(section.flatten())
            >>> print(model.field_a)
            value

            ```

        Serialize a model to an INI format:
            ```python
            >>> from hydrolib.core.dflowfm.ini.serializer import INISerializerConfig
            >>> from hydrolib.core.base.models import ModelSaveSettings
            >>> config = INISerializerConfig()
            >>> section = model._to_section(config, save_settings=ModelSaveSettings())
            >>> print(section.header)
            MyHeader

            ```

    Notes:
        - Subclasses can override the `_header` attribute to define the INI block header.
        - Arbitrary fields can be added dynamically and are included during serialization.
    """

    _header: str = ""
    _file_path_style_converter = FilePathStyleConverter()
    model_config = ConfigDict(
        extra="ignore",
        arbitrary_types_allowed=False,
        populate_by_name=True,
        error_url_template=None,
    )

    @classmethod
    def _get_unknown_keyword_error_manager(cls) -> Optional[UnknownKeywordErrorManager]:
        """
        Retrieves the error manager for handling unknown keywords in INI files.

        Returns:
            Optional[UnknownKeywordErrorManager]:
                An instance of the error manager or None if unknown keywords are allowed.
        """
        return UnknownKeywordErrorManager()

    @classmethod
    def _supports_comments(cls) -> bool:
        """
        Indicates whether the model supports comments for its fields.

        Returns:
            bool: True if comments are supported; otherwise, False.
        """
        return True

    @classmethod
    def _duplicate_keys_as_list(cls) -> bool:
        """
        Indicates whether duplicate keys in INI sections should be treated as lists.

        Returns:
            bool: True if duplicate keys should be treated as lists; otherwise, False.
        """
        return False

    @classmethod
    def get_list_delimiter(cls) -> str:
        """List delimiter string that will be used for serializing
        list field values for any IniBasedModel, **if** that field has
        no custom list delimiter.

        This function should be overridden by any subclass for a particular
        filetype that needs a specific/different list separator.
        """
        return " "

    @classmethod
    def get_list_field_delimiter(cls, field_key: str) -> str:
        """List delimiter string that will be used for serializing
        the given field's value.
        The returned delimiter is either the field's custom list delimiter
        if that was specified using Field(.., delimiter=".."), or the
        default list delimiter for the model class that this field belongs
        to.

        Args:
            field_key (str): the original field key (not its alias).

        Returns:
            str: the delimiter string to be used for serializing the given field.
        """
        delimiter = None
        field_info = cls.model_fields.get(field_key)
        if (
            (field := field_info)
            and isinstance(field, FieldInfo)
            and field.json_schema_extra
        ):
            delimiter = field.json_schema_extra.get("delimiter")
        if not delimiter:
            delimiter = cls.get_list_delimiter()

        return delimiter

    class Comments(BaseModel, ABC):
        """Represents the comments associated with fields in an INIBasedModel.

        Config:
            extra: Extra.allow
                Allows dynamic fields for comments.
            arbitrary_types_allowed: bool
                Indicates that only known types are allowed.
        """

        model_config = ConfigDict(extra="allow", arbitrary_types_allowed=False)

    comments: Optional[Comments] = Comments()

    @model_validator(mode="before")
    @classmethod
    def _validate_unknown_keywords(cls, values):
        """Validates fields and raises errors for unknown keywords.

        Args:
            values (dict): Dictionary of field values to validate.

        Returns:
            dict: Validated field values.

        Raises:
            ValueError: If unknown keywords are found.
        """
        unknown_keyword_error_manager = cls._get_unknown_keyword_error_manager()
        do_not_validate = cls._exclude_from_validation(values)
        if unknown_keyword_error_manager:
            unknown_keyword_error_manager.raise_error_for_unknown_keywords(
                values,
                cls._header,
                cls.model_fields,
                cls._exclude_fields() | do_not_validate,
            )
        return values

    @model_validator(mode="before")
    @classmethod
    def _skip_nones_and_set_header(cls, values):
        """Drop None fields for known fields.

        Filters out None values and sets the model header.

        Args:
            values (dict): Dictionary of field values.

        Returns:
            dict: Updated field values with None values removed.
        """
        # Convert Section to dictionary if needed
        values = cls._convert_section_to_dict(values)
        if not isinstance(values, dict):
            return values

        dropkeys = []
        for k, v in values.items():
            if v is None and k in cls.model_fields.keys():
                dropkeys.append(k)

        logger.info(f"Dropped unset keys: {dropkeys}")
        for k in dropkeys:
            values.pop(k)

        if "_header" in values:
            values["_header"] = cls._header

        return values

    @field_validator("comments", mode="after")
    @classmethod
    def comments_matches_has_comments(cls, v):
        """
        Validates the presence of comments if supported by the model.

        Args:
            v (Any): The comments field value.

        Returns:
            Any: Validated comments field value.
        """
        if not cls._supports_comments() and v is not None:
            logging.warning(f"Dropped unsupported comments from {cls.__name__} init.")
            v = None
        return v

    def __setattr__(self, key, value) -> None:
        """
        Custom setter to handle Fortran-style scientific notation conversion
        for float fields when setting attributes.
        """
        field = self.__class__.model_fields.get(key)
        if field is None:
            super().__setattr__(key, value)
            return

        field_type = field.annotation
        if field_type in (float, List[float]):
            value = FortranScientificNotationConverter.convert(value)
        super().__setattr__(key, value)

    @classmethod
    def preprocess_input(cls, values: dict) -> dict:
        """Pre-process input values for the model.

        The preprocess_input method is called before validation and triggers the following actions:
            - Convert Fortran-style scientific notation to Python float.
        """
        if isinstance(values, dict):
            new_values = FortranScientificNotationConverter.convert_fields(
                values, cls.model_fields
            )
        else:
            new_values = values

        return new_values

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        schema = handler(source_type)

        return core_schema.no_info_before_validator_function(
            cls.preprocess_input,  # this is called BEFORE validation
            schema,
        )

    @classmethod
    def _convert_section_to_dict(cls, value: Any) -> Any:
        """
        Converts a Section object to a dictionary if needed.

        Args:
            value (Any): The value to convert.

        Returns:
            Any: The converted value (dictionary) or the original value if not a Section.
        """
        if isinstance(value, Section):
            return value.flatten(
                cls._duplicate_keys_as_list(), cls._supports_comments()
            )
        return value

    @classmethod
    def _extract_file_model_from_section(
        cls, section: Section, key: str, file_model_class: Type
    ):
        """Extracts a file model from a Section object.

        Args:
            section (Section):
                The Section object to extract from.
            key (str):
                The key to look for in the Section.
            file_model_class (Type):
                The class to use for creating the file model.

        Returns:
            Optional[Any]:
                The file model if found, None otherwise.
        """
        for prop in section.content:
            if isinstance(prop, Property) and prop.key.lower() == key.lower():
                return file_model_class(prop.value)
        return None

    @classmethod
    def model_validate(cls: Type["INIBasedModel"], value: Any) -> "INIBasedModel":
        """
        Validates a value as an instance of INIBasedModel.

        Args:
            value (Any): The value to validate.

        Returns:
            INIBasedModel: The validated instance.
        """
        value = cls._convert_section_to_dict(value)
        return super().model_validate(value)

    @classmethod
    def _exclude_from_validation(cls, input_data: Optional[dict] = None) -> Set:
        """
        Fields that should not be checked when validating existing fields as they will be dynamically added.

        Args:
            input_data (Optional[dict]): Input data to process.

        Returns:
            Set: Set of field names to exclude from validation.
        """
        return set()

    @classmethod
    def _exclude_fields(cls) -> Set:
        """
        Defines fields to exclude from serialization.

        Returns:
            Set: Set of field names to exclude.
        """
        return {"comments", "datablock", "_header"}

    def _convert_value(
        self,
        key: str,
        v: Any,
        config: INISerializerConfig,
        save_settings: ModelSaveSettings,
    ) -> str:
        """
        Converts a field value to its serialized string representation.

        Args:
            key (str): The field key.
            v (Any): The field value.
            config (INISerializerConfig): Configuration for serialization.
            save_settings (ModelSaveSettings): Settings for saving the model.

        Returns:
            str: The serialized value.
        """
        if isinstance(v, bool):
            return str(int(v))
        elif isinstance(v, list):
            convert_value = lambda x: self._convert_value(key, x, config, save_settings)
            return self.__class__.get_list_field_delimiter(key).join(
                [convert_value(x) for x in v]
            )
        elif isinstance(v, Enum):
            return v.value
        elif isinstance(v, float):
            return f"{v:{config.float_format}}"
        elif isinstance(v, FileModel) and v.filepath is not None:
            return self._file_path_style_converter.convert_from_os_style(
                v.filepath, save_settings.path_style
            )
        elif v is None:
            return ""
        else:
            return str(v)

    def _to_section(
        self, config: INISerializerConfig, save_settings: ModelSaveSettings
    ) -> Section:
        """
        Converts the model to an INI section.

        Args:
            config (INISerializerConfig): Configuration for serialization.
            save_settings (ModelSaveSettings): Settings for saving the model.

        Returns:
            Section: The INI section representation of the model.
        """
        props = []
        cls_fields = type(self).model_fields  # cache the class-level dict
        for key, value in self:
            if not self._should_be_serialized(key, value, save_settings):
                continue

            field_key = key
            if key in cls_fields:
                key = cls_fields[key].alias

            prop = Property(
                key=key,
                value=self._convert_value(field_key, value, config, save_settings),
                comment=getattr(self.comments, field_key, None),
            )
            props.append(prop)
        return Section(header=self._header, content=props)

    def _should_be_serialized(
        self, key: str, value: Any, save_settings: ModelSaveSettings
    ) -> bool:
        """
        Determines if a field should be serialized.

        Args:
            key (str): The field key.
            value (Any): The field value.
            save_settings (ModelSaveSettings): Settings for saving the model.

        Returns:
            bool: True if the field should be serialized; otherwise, False.
        """
        if key in self._exclude_fields():
            return False

        if save_settings._exclude_unset and key not in self.model_fields_set:
            return False

        cls_fields = type(self).model_fields
        field = cls_fields.get(key)
        if not field:
            return value is not None

        field_type = field.annotation
        if self._is_union(field_type):
            return value is not None or self._union_has_filemodel(field_type)

        if self._is_list(field_type):
            field_type = get_args(field_type)[0]

        return self._value_is_not_none_or_type_is_filemodel(field_type, value)

    @staticmethod
    def _is_union(field_type: type) -> bool:
        """
        Checks if a type is a Union.

        Args:
            field_type (type): The type to check.

        Returns:
            bool: True if the type is a Union; otherwise, False.
        """
        return get_origin(field_type) is Union

    @staticmethod
    def _union_has_filemodel(field_type: type) -> bool:
        """
        Checks if a Union type includes a FileModel subtype.

        Args:
            field_type (type): The type to check.

        Returns:
            bool: True if the Union includes a FileModel; otherwise, False.
        """
        return any(
            isclass(arg) and issubclass(arg, FileModel) for arg in get_args(field_type)
        )

    @staticmethod
    def _is_list(field_type: type) -> bool:
        """
        Checks if a type is a list.

        Args:
            field_type (type): The type to check.

        Returns:
            bool: True if the type is a list; otherwise, False.
        """
        return get_origin(field_type) is List or get_origin(field_type) is list

    @staticmethod
    def _value_is_not_none_or_type_is_filemodel(field_type: type, value: Any) -> bool:
        """
        Checks if a value is not None or if its type is FileModel.

        Args:
        field_type (type): The expected type of the field.
        value (Any): The value to check.

        Returns:
            bool: True if the value is valid; otherwise, False.
        """
        return value is not None or issubclass(field_type, FileModel)


Datablock = List[List[Union[float, str]]]


class DataBlockINIBasedModel(INIBasedModel):
    """DataBlockINIBasedModel defines the base model for ini models with datablocks.

    This class extends the functionality of INIBasedModel to handle structured data blocks
    commonly found in INI files. It provides validation, serialization, and conversion methods
    for working with these data blocks.

    Attributes:
        datablock (Datablock): (class attribute) the actual data columns.

    Attributes:
        datablock (List[List[Union[float, str]]], optional):
            The initial data block for the model. Defaults to an empty list.

    Raises:
        ValueError: If a NaN value is found within the data block.

    See Also:
        INIBasedModel: The parent class for models representing INI-based configurations.
        INISerializerConfig: Provides configuration for INI serialization.

    Examples:
        Create a model and validate its data block:
            ```python
            >>> from hydrolib.core.dflowfm.ini.models import DataBlockINIBasedModel
            >>> model = DataBlockINIBasedModel(datablock=[[1.0, 2.0], [3.0, 4.0]])
            >>> print(model.datablock)
            [[1.0, 2.0], [3.0, 4.0]]

            ```

        Attempt to create a model with invalid data:
            ```python
            >>> try:
            ...     model = DataBlockINIBasedModel(datablock=[[1.0, None]])
            ... except Exception as e:
            ...     print(e)
            1 validation error for DataBlockINIBasedModel
            datablock -> 0 -> 1
              none is not an allowed value (type=type_error.none.not_allowed)

            ```

    Notes:
        - The class includes a validator to ensure that no NaN values are present in the data block.
        - Data blocks are converted to a serialized format for writing to INI files.
    """

    datablock: Annotated[Datablock, BeforeValidator(make_list)] = Field(
        default_factory=list
    )

    @staticmethod
    def _is_float(value: str) -> bool:
        """
        Determines whether a value is a float.

        Args:
            value (str): The value to check.

        Returns:
            bool: True if the value can be converted to a float; otherwise, False.
        """
        try:
            float(value)
            return True
        except ValueError:
            return False

    @field_validator("datablock", mode="before")
    @classmethod
    def _convert_to_float(cls, datablock: Datablock) -> Datablock:
        """
        Validates that all values in the datablock are either floats or strings.

        Args:
            datablock (Datablock): The datablock to validate.

        Returns:
            Datablock: The validated datablock.

        Raises:
            ValueError: If any value in the datablock is not a float or string.
        """
        for r, row in enumerate(datablock):
            for c, value in enumerate(row):
                if isinstance(value, str) and cls._is_float(value):
                    datablock[r][c] = float(value)
        return datablock

    @classmethod
    def _get_unknown_keyword_error_manager(cls) -> Optional[UnknownKeywordErrorManager]:
        """
        The DataBlockINIBasedModel does not need to raise an error on unknown keywords.

        Returns:
            Optional[UnknownKeywordErrorManager]: Returns None as unknown keywords are ignored.
        """
        return None

    def as_dataframe(self) -> DataFrame:
        """Convert the datablock as a pandas DataFrame

        - The first number from each list in the block as an index for that row.

        Returns:
            DataFrame: The datablock as a pandas DataFrame.

        Examples:
                >>> from hydrolib.core.dflowfm.ini.models import DataBlockINIBasedModel
                >>> model = DataBlockINIBasedModel(datablock=[[0, 10, 100], [1, 20, 200]])
                >>> df = model.as_dataframe()
                >>> print(df)
                        0      1
                0.0  10.0  100.0
                1.0  20.0  200.0
        """
        df = DataFrame(self.datablock).set_index(0)
        df.index.name = None
        df.columns = range(len(df.columns))
        return df

    def _to_section(
        self,
        config: DataBlockINIBasedSerializerConfig,
        save_settings: ModelSaveSettings,
    ) -> Section:
        """
        Converts the current model to an INI Section representation.

        Args:
            config (DataBlockINIBasedSerializerConfig): Configuration for serializing the data block.
            save_settings (ModelSaveSettings): Settings for saving the model.

        Returns:
            Section: The INI Section containing serialized data and the data block.
        """
        section = super()._to_section(config, save_settings)
        section.datablock = self._to_datablock(config)
        return section

    def _to_datablock(self, config: DataBlockINIBasedSerializerConfig) -> List[List]:
        """
        Converts the data block to a serialized format based on the configuration.

        Args:
            config (DataBlockINIBasedSerializerConfig): Configuration for serializing the data block.

        Returns:
            List[List]: A serialized representation of the data block.
        """
        converted_datablock = []

        for row in self.datablock:
            converted_row = (
                DataBlockINIBasedModel.convert_value(value, config) for value in row
            )
            converted_datablock.append(list(converted_row))

        return converted_datablock

    @classmethod
    def convert_value(
        cls, value: Union[float, str], config: DataBlockINIBasedSerializerConfig
    ) -> str:
        """
        Converts a value in the data block to its serialized string representation.

        Args:
            value (Union[float, str]): The value to be converted.
            config (DataBlockINIBasedSerializerConfig): Configuration for the conversion.

        Returns:
            str: The serialized string representation of the value.
        """
        if isinstance(value, float):
            return f"{value:{config.float_format_datablock}}"

        return value

    @field_validator("datablock", mode="after")
    @classmethod
    def _validate_no_nans_are_present(cls, datablock: Datablock) -> Datablock:
        """Validate that the datablock does not have any NaN values.

        Args:
            datablock (Datablock): The datablock to validate.

        Raises:
            ValueError: When a NaN is present in the datablock.

        Returns:
            Datablock: The validated datablock.
        """
        if any(cls._is_float_and_nan(value) for row in datablock for value in row):
            raise ValueError("NaN is not supported in datablocks.")

        return datablock

    @staticmethod
    def _is_float_and_nan(value: float) -> bool:
        """
        Determines whether a value is a float and is NaN.

        Args:
            value (float): The value to check.

        Returns:
            bool: True if the value is a NaN float; otherwise, False.
        """
        return isinstance(value, float) and isnan(value)


class INIGeneral(INIBasedModel):
    _header: Literal["General"] = "General"
    fileversion: str = Field("3.00", alias="fileVersion")
    filetype: str = Field(alias="fileType")

    @classmethod
    def _supports_comments(cls):
        return True


class INIModel(ParsableFileModel):
    """INI Model representation of a *.ini file.

    Typically subclasses will implement the various sorts of ini files,
    specifically for their fileType/contents.
    Child elements of this class associated with chapters/blocks in the
    ini file will be (sub)class of INIBasedModel.
    """

    serializer_config: INISerializerConfig = INISerializerConfig()

    general: INIGeneral

    @classmethod
    def _ext(cls) -> str:
        return ".ini"

    @classmethod
    def _filename(cls) -> str:
        return "fm"

    @classmethod
    def _get_serializer(cls):
        pass  # unused in favor of direct _serialize

    @classmethod
    def _get_parser(cls) -> Callable:
        return Parser.parse_as_dict

    def _to_document(self, save_settings: ModelSaveSettings) -> Document:
        header = CommentBlock(lines=[f"written by HYDROLIB-core {version}"])
        sections = []
        for key, value in self:
            if key in self._exclude_fields() or value is None:
                continue
            if save_settings._exclude_unset and key not in self.model_fields_set:
                continue
            if isinstance(value, list):
                for v in value:
                    sections.append(
                        v._to_section(self.serializer_config, save_settings)
                    )
            else:
                sections.append(
                    value._to_section(self.serializer_config, save_settings)
                )
        return Document(header_comment=[header], sections=sections)

    def _serialize(self, _: dict, save_settings: ModelSaveSettings) -> None:
        """
        Create a `Document` from the model and write it to the file.
        """
        write_ini(
            self._resolved_filepath,
            self._to_document(save_settings),
            config=self.serializer_config,
        )



from typing import Any, Dict, List, Literal, Optional

from pydantic import Field, ValidationInfo, field_validator

from hydrolib.core.base.models import DiskOnlyFileModel
from hydrolib.core.base.utils import PathToDictionaryConverter
from hydrolib.core.dflowfm.ini.models import INIBasedModel
from hydrolib.core.dflowfm.ini.util import split_string_on_delimiter
from hydrolib.core.dflowfm.mdu import (
    FMModel,
    General,
    Geometry,
    Numerics,
    Output,
    Physics,
    Processes,
    Restart,
    Sediment,
    Time,
    Trachytopes,
    Waves,
    Wind,
)
from hydrolib.core.dflowfm.polyfile import PolyFile

DEPRECATED_VARIABLE = "Deprecated variable."
DEPRECATED_KEYWORD = "Deprecated keyword."


class ResearchGeneral(General):
    """An extended [general] section that includes highly experimental research keywords."""

    class Comments(General.Comments):
        research_modelspecific: Optional[str] = Field(
            "Optional 'model specific ID', to enable certain custom runtime function calls (instead of via MDU name).",
            alias="modelspecific",
        )
        research_inputspecific: Optional[str] = Field(
            "Use of hardcoded specific inputs, shall not be used by users (0: no, 1: yes).",
            alias="inputspecific",
        )

    comments: Comments = Comments()

    research_modelspecific: Optional[str] = Field(None, alias="modelspecific")
    research_inputspecific: Optional[bool] = Field(None, alias="inputspecific")


class ResearchGeometry(Geometry):
    """An extended [geometry] section that includes highly experimental research keywords."""

    class Comments(Geometry.Comments):
        research_toplayminthick: Optional[str] = Field(
            "Minimum top layer thickness(m), only for Z-layers.",
            alias="toplayminthick",
        )
        research_helmert: Optional[str] = Field(
            "Use Helmert (0: no, 1: yes).",
            alias="helmert",
        )
        research_waterdepthini1d: Optional[str] = Field(
            "Initial waterdepth in 1D.",
            alias="waterdepthini1d",
        )
        research_zlayeratubybob: Optional[str] = Field(
            "Lowest connected cells governed by bob instead of by bL L/R.",
            alias="zlayeratubybob",
        )
        research_shipdeffile: Optional[str] = Field(
            "File *.shd containing ship definitions.",
            alias="shipdeffile",
        )
        research_stripmesh: Optional[str] = Field(
            "Strip unused nodes and links from the mesh after clipping (1: strip, 0: do not strip).",
            alias="stripmesh",
        )
        research_bedwavelength: Optional[str] = Field(
            "Bed testcases.",
            alias="bedwavelength",
        )
        research_removesmalllinkstrsh: Optional[str] = Field(
            "0.1	0-1, 0= no removes.",
            alias="removesmalllinkstrsh",
        )
        research_createlinks1d2d: Optional[str] = Field(
            "Ruecksichtslos create links between 1D nodes and 2D cells when initializing model (1: yes, 0: no).",
            alias="createlinks1d2d",
        )
        research_bedwaveamplitude: Optional[str] = Field(
            "Bed testcases.",
            alias="bedwaveamplitude",
        )
        research_uniformhu: Optional[str] = Field(
            "Waterdepth in rigid-lid-like solution.",
            alias="uniformhu",
        )
        research_tsigma: Optional[str] = Field(
            "Sigma Adaptation period for Layertype==4 (s).",
            alias="tsigma",
        )
        research_dpuopt: Optional[str] = Field(
            "Bed level interpolation at velocity point in case of tile approach bed level: 1 = max (default); 2 = mean.",
            alias="dpuopt",
        )
        research_ihuzcsig: Optional[str] = Field(
            "If keepzlayeringatbed>=2 : 1,2,3=av,mx,mn of Leftsig,Rightsig,4=uniform'",
            alias="ihuzcsig",
        )
        research_ihuz: Optional[str] = Field(
            "ihuz",
            alias="ihuz",
        )
        research_cosphiutrsh: Optional[str] = Field(
            "0-1, 1= no bad orthos.",
            alias="cosphiutrsh",
        )
        research_cutcelllist: Optional[str] = Field(
            "File with names of cutcell polygons, e.g. cutcellpolygons.lst.",
            alias="cutcelllist",
        )
        research_uniformtyp1d: Optional[str] = Field(
            "Uniform type for channel profiles not specified by profloc.",
            alias="uniformtyp1d",
        )
        research_oned2dinternallinktype: Optional[str] = Field(
            "Link treatment method for type-3 internal links.",
            alias="oned2dinternallinktype",
        )
        research_orgfloorlevtoplaydef: Optional[str] = Field(
            "Keep original definition of Floor level of top layer.",
            alias="orgfloorlevtoplaydef",
        )
        research_pipefile: Optional[str] = Field(
            "File *.pliz containing pipe-based 'culverts'.",
            alias="pipefile",
        )
        research_groundlayerthickness: Optional[str] = Field(
            "Only in pipes: groundlayer thickness (m).",
            alias="groundlayerthickness",
        )
        research_extrbl: Optional[str] = Field(
            "Extrapolation of bed level at boundaries according to the slope: 0 = no extrapolation (default); 1 = extrapolate.",
            alias="extrbl",
        )
        research_keepzlay1bedvol: Optional[str] = Field(
            "Correct volumes when keepzlayeringatbed=1 (0: too large bedcell volumes, 1: correct bedcell volumes).",
            alias="keepzlay1bedvol",
        )
        research_stretchcoef: Optional[str] = Field(
            "Layers thickness percentage",
            alias="stretchcoef",
        )

    comments: Comments = Comments()

    research_toplayminthick: Optional[float] = Field(None, alias="toplayminthick")
    research_helmert: Optional[bool] = Field(None, alias="helmert")
    research_waterdepthini1d: Optional[float] = Field(None, alias="waterdepthini1d")
    research_zlayeratubybob: Optional[int] = Field(None, alias="zlayeratubybob")
    research_shipdeffile: Optional[DiskOnlyFileModel] = Field(None, alias="shipdeffile")
    research_stripmesh: Optional[bool] = Field(None, alias="stripmesh")
    research_bedwavelength: Optional[float] = Field(None, alias="bedwavelength")
    research_removesmalllinkstrsh: Optional[float] = Field(
        None, alias="removesmalllinkstrsh"
    )
    research_createlinks1d2d: Optional[bool] = Field(None, alias="createlinks1d2d")
    research_bedwaveamplitude: Optional[float] = Field(None, alias="bedwaveamplitude")
    research_uniformhu: Optional[float] = Field(None, alias="uniformhu")
    research_tsigma: Optional[float] = Field(None, alias="tsigma")
    research_dpuopt: Optional[int] = Field(None, alias="dpuopt")
    research_ihuzcsig: Optional[int] = Field(None, alias="ihuzcsig")
    research_ihuz: Optional[int] = Field(None, alias="ihuz")
    research_cosphiutrsh: Optional[float] = Field(None, alias="cosphiutrsh")
    research_cutcelllist: Optional[DiskOnlyFileModel] = Field(None, alias="cutcelllist")
    research_uniformtyp1d: Optional[int] = Field(None, alias="uniformtyp1d")
    research_oned2dinternallinktype: Optional[int] = Field(
        None, alias="oned2dinternallinktype"
    )
    research_orgfloorlevtoplaydef: Optional[bool] = Field(
        None, alias="orgfloorlevtoplaydef"
    )
    research_pipefile: Optional[DiskOnlyFileModel] = Field(None, alias="pipefile")
    research_groundlayerthickness: Optional[float] = Field(
        None, alias="groundlayerthickness"
    )
    research_extrbl: Optional[bool] = Field(None, alias="extrbl")
    research_keepzlay1bedvol: Optional[bool] = Field(None, alias="keepzlay1bedvol")
    research_stretchcoef: Optional[str] = Field(None, alias="stretchcoef")


class ResearchNumerics(Numerics):
    """An extended [numerics] section that includes highly experimental research keywords."""

    class Comments(Numerics.Comments):
        research_faclaxturb: Optional[str] = Field(
            "Default: 0=TurKin0 from links, 1.0=from nodes. 0.5=fityfifty.",
            alias="faclaxturb",
        )
        research_jafaclaxturbtyp: Optional[str] = Field(
            "Vertical distr of facLaxturb, 1=: (sigm<0.5=0.0 sigm>0.75=1.0 linear in between), 2:=1.0 for whole column.",
            alias="jafaclaxturbtyp",
        )
        research_locsaltmin: Optional[str] = Field(
            "Minimum salinity for case of lock exchange.",
            alias="locsaltmin",
        )
        research_lincontin: Optional[str] = Field(
            "Default 0; Set to 1 for linearizing d(Hu)/dx; link to AdvecType.",
            alias="lincontin",
        )
        research_cfexphormom: Optional[str] = Field(
            "Exponent for including (1-CFL) in HO term horizontal mom.",
            alias="cfexphormom",
        )
        research_jbasqbnddownwindhs: Optional[str] = Field(
            "Water depth scheme at discharge boundaries (0: original hu, 1: downwind hs).",
            alias="jbasqbnddownwindhs",
        )
        research_coriohhtrsh: Optional[str] = Field(
            "0=default=no safety in hu/hus weightings, only for Newcorio=1.",
            alias="coriohhtrsh",
        )
        research_limtypw: Optional[str] = Field(
            "Limiter type for wave action transport (0: none, 1: minmod, 2: van Leer, 3: Koren, 4: monotone central).",
            alias="limtypw",
        )
        research_huweirregular: Optional[str] = Field(
            "For villemonte and Tabellenboek, regular hu below Huweirregular.",
            alias="huweirregular",
        )
        research_structurelayersactive: Optional[str] = Field(
            "0=structure flow through all layers, 1=structure flow only through open layers.",
            alias="structurelayersactive",
        )
        research_corioadamsbashfordfac: Optional[str] = Field(
            "Adams-Bashford factor in Coriolis term (0: No/explicit, 0.5d0=Adams-Bashford), only for Newcorio=1.",
            alias="corioadamsbashfordfac",
        )
        research_baorgfracmin: Optional[str] = Field(
            "Cell area = max(orgcellarea*Baorgfracmin, cutcell area)",
            alias="baorgfracmin",
        )
        research_epstke: Optional[str] = Field(
            "TKE=max(TKE, EpsTKE), default=1d-32", alias="epstke"
        )
        research_jadrhodz: Optional[str] = Field(
            "1:central org, 2:centralnew, 3:upw cell, 4:most stratf. cell, 5:least stratf. cell.",
            alias="jadrhodz",
        )
        research_logprofkepsbndin: Optional[str] = Field(
            "Inflow: 0=0 keps, 1 = log keps inflow, 2 = log keps in and outflow.",
            alias="logprofkepsbndin",
        )
        research_epshstem: Optional[str] = Field(
            "Only compute heatflx + evap if depth > epshstem.", alias="epshstem"
        )
        research_newcorio: Optional[str] = Field(
            "0=prior to 27-11-2019, 1=no normal forcing on open bnds, plus 12 variants.",
            alias="newcorio",
        )
        research_diffusiononbnd: Optional[str] = Field(
            "On open boundaries, 0 switches off horizontal diffusion Default = 1.",
            alias="diffusiononbnd",
        )
        research_barrieradvection: Optional[str] = Field(
            "1 = no correction, 2 = advection correction.", alias="barrieradvection"
        )
        research_rhointerfaces: Optional[str] = Field(
            "Evaluate rho at interfaces, 0=org at centers, 1=at interfaces.",
            alias="rhointerfaces",
        )
        research_logprofatubndin: Optional[str] = Field(
            "ubnds inflow: 0=uniform U1, 1 = log U1, 2 = user3D.",
            alias="logprofatubndin",
        )
        research_horadvtypzlayer: Optional[str] = Field(
            "Vertical treatment of horizontal advection in z-layers (0: default, 1: N/A, 2: sigma-like).",
            alias="horadvtypzlayer",
        )
        research_chkdifd: Optional[str] = Field(
            "Check diffusion terms if depth < chkdifd, only if jatransportautotimestepdiff==1.",
            alias="chkdifd",
        )
        research_fixedweirfrictscheme: Optional[str] = Field(
            "Fixed weir friction scheme (0: friction based on hu, 1: friction based on subgrid weir friction scheme).",
            alias="fixedweirfrictscheme",
        )
        research_icoriolistype: Optional[str] = Field(
            "0=No, 5=default, 3, 4 no weights, 5-10 Kleptsova hu/hs, 25-30 Ham hs/hu, odd: 2D hs/hu, even: hsk/huk.",
            alias="icoriolistype",
        )
        research_zwsbtol: Optional[str] = Field(
            "Tolerance for zws(kb-1) at bed.", alias="zwsbtol"
        )
        research_cfexphu: Optional[str] = Field(
            "Exp for including (1-CFL) in sethu.", alias="cfexphu"
        )
        research_drop3d: Optional[str] = Field(
            "Waterdepth factor, apply droplosses in 3D if z upwind below bob + 2/3 hu*drop3D.",
            alias="drop3d",
        )
        research_zlayercenterbedvel: Optional[str] = Field(
            "Reconstruction of center velocity at half closed bedcells (0=no, 1: copy bed link velocities).",
            alias="zlayercenterbedvel",
        )
        research_cffacver: Optional[str] = Field(
            "Factor for including (1-CFL) in HO term vertical (0d0: no, 1d0: yes).",
            alias="cffacver",
        )
        research_eddyviscositybedfacmax: Optional[str] = Field(
            "Limit eddy viscosity at bed (factor 0.0-1.0 of first layer above).",
            alias="eddyviscositybedfacmax",
        )
        research_epseps: Optional[str] = Field(
            "EPS=max(EPS, EpsEPS), default=1d-32, (or TAU).", alias="epseps"
        )
        research_lateral_fixedweir_umin_method: Optional[str] = Field(
            "Method for minimal velocity treshold for weir losses in iterative lateral 1d2d weir coupling.",
            alias="lateral_fixedweir_umin_method",
        )
        research_lateral_fixedweir_minimal_1d2d_embankment: Optional[str] = Field(
            "Minimal crest height of 1D2D SOBEK-DFM embankments.",
            alias="lateral_fixedweir_minimal_1d2d_embankment",
        )
        research_testfixedweirs: Optional[str] = Field(
            "Test for fixed weir algoritms (0 = Sieben2010, 1 = Sieben2007 ).",
            alias="testfixedweirs",
        )
        research_jposhchk: Optional[str] = Field(
            "Check for positive waterdepth (0: no, 1: 0.7dts, just redo, 2: 1.0dts, close all links, 3: 0.7dts, close all links, 4: 1.0dts, reduce au, 5: 0.7dts, reduce au, 6: 1.0dts, close outflowing links, 7: 0.7*dts, close outflowing link.",
            alias="jposhchk",
        )
        research_cfconhormom: Optional[str] = Field(
            "Constant for including (1-CFL) in HO term horizontal mom.",
            alias="cfconhormom",
        )
        research_cffachormom: Optional[str] = Field(
            "Factor for including (1-CFL) in HO term horizontal mom (0d0: no, 1d0: yes).",
            alias="cffachormom",
        )
        research_trsh_u1lb: Optional[str] = Field(
            "2D bedfriction in 3D below this threshold (m).", alias="trsh_u1lb"
        )
        research_corioconstant: Optional[str] = Field(
            "0=default, 1=Coriolis constant in sferic models anyway, 2=beta plane, both in cart. and spher. coord.",
            alias="corioconstant",
        )
        research_jaupwindsrc: Optional[str] = Field(
            "1st-order upwind advection at sources/sinks (1) or higher-order (0).",
            alias="jaupwindsrc",
        )
        research_locsaltlev: Optional[str] = Field(
            "Salinity level for case of lock exchange.", alias="locsaltlev"
        )
        research_subsuplupdates1: Optional[str] = Field(
            "Update water levels (S1) due to subsidence / uplift.",
            alias="subsuplupdates1",
        )
        research_linkdriedmx: Optional[str] = Field(
            "Nr of Au reduction steps after having dried.", alias="linkdriedmx"
        )
        research_maxitpresdens: Optional[str] = Field(
            "Max nr of iterations in pressure-density coupling, only used if idensform > 10.",
            alias="maxitpresdens",
        )
        research_lateral_fixedweir_relax: Optional[str] = Field(
            "Relaxation factor for iterative lateral 1d2d weir coupling algorithm.",
            alias="lateral_fixedweir_relax",
        )
        research_numlimdt_baorg: Optional[str] = Field(
            "If previous numlimdt > Numlimdt_baorg keep original cell area ba in cutcell.",
            alias="numlimdt_baorg",
        )
        research_locsaltmax: Optional[str] = Field(
            "Maximum salinity for case of lock exchange.",
            alias="locsaltmax",
        )
        research_cffachu: Optional[str] = Field(
            "Factor for including (1-CFL) in sethu (0d0: no, 1d0: yes).",
            alias="cffachu",
        )
        research_vertadvtypmom3onbnd: Optional[str] = Field(
            "Vert. adv. u1 bnd UpwimpL: 0=follow javau , 1 = on bnd, 2= on and near bnd.",
            alias="vertadvtypmom3onbnd",
        )
        research_noderivedtypes: Optional[str] = Field(
            "0=use der. types. , 1 = less, 2 = lesser, 5 = also dealloc der. types.",
            alias="noderivedtypes",
        )

        research_jarhoxu: Optional[str] = Field(
            "Include density gradient in advection term (0: no(strongly advised), 1: yes, 2: Also in barotropic and baroclinic pressure term, 3,4: Also in vertical advection).",
            alias="jarhoxu",
        )
        research_jaorgsethu: Optional[str] = Field(
            DEPRECATED_VARIABLE,
            alias="jaorgsethu",
        )
        research_cflwavefrac: Optional[str] = Field(
            DEPRECATED_VARIABLE,
            alias="cflwavefrac",
        )
        research_jaembed1d: Optional[str] = Field(
            DEPRECATED_VARIABLE,
            alias="jaembed1d",
        )
        research_maxitverticalforester: Optional[str] = Field(
            "Forester iterations for salinity (0: no vertical filter for salinity, > 0: max nr of iterations).",
            alias="maxitverticalforester",
        )

        research_ilutype: Optional[str] = Field(
            "0: parms-default",
            alias="ilutype",
        )
        research_nlevel: Optional[str] = Field(
            "0d0: parms-default",
            alias="nlevel",
        )
        research_dtol: Optional[str] = Field(
            "0d0: parms-default",
            alias="dtol",
        )

    comments: Comments = Comments()

    research_faclaxturb: Optional[float] = Field(None, alias="faclaxturb")
    research_jafaclaxturbtyp: Optional[int] = Field(None, alias="jafaclaxturbtyp")
    research_locsaltmin: Optional[float] = Field(None, alias="locsaltmin")
    research_lincontin: Optional[int] = Field(None, alias="lincontin")
    research_cfexphormom: Optional[float] = Field(None, alias="cfexphormom")
    research_jbasqbnddownwindhs: Optional[int] = Field(None, alias="jbasqbnddownwindhs")
    research_coriohhtrsh: Optional[float] = Field(None, alias="coriohhtrsh")
    research_limtypw: Optional[int] = Field(None, alias="limtypw")
    research_huweirregular: Optional[float] = Field(None, alias="huweirregular")
    research_structurelayersactive: Optional[int] = Field(
        None, alias="structurelayersactive"
    )
    research_corioadamsbashfordfac: Optional[float] = Field(
        None, alias="corioadamsbashfordfac"
    )
    research_baorgfracmin: Optional[float] = Field(None, alias="baorgfracmin")
    research_epstke: Optional[float] = Field(None, alias="epstke")
    research_jadrhodz: Optional[int] = Field(None, alias="jadrhodz")
    research_logprofkepsbndin: Optional[int] = Field(None, alias="logprofkepsbndin")
    research_epshstem: Optional[float] = Field(None, alias="epshstem")
    research_newcorio: Optional[bool] = Field(None, alias="newcorio")
    research_diffusiononbnd: Optional[bool] = Field(None, alias="diffusiononbnd")
    research_barrieradvection: Optional[int] = Field(None, alias="barrieradvection")
    research_rhointerfaces: Optional[int] = Field(None, alias="rhointerfaces")
    research_logprofatubndin: Optional[int] = Field(None, alias="logprofatubndin")
    research_horadvtypzlayer: Optional[int] = Field(None, alias="horadvtypzlayer")
    research_chkdifd: Optional[float] = Field(None, alias="chkdifd")
    research_fixedweirfrictscheme: Optional[int] = Field(
        None, alias="fixedweirfrictscheme"
    )
    research_icoriolistype: Optional[int] = Field(None, alias="icoriolistype")
    research_zwsbtol: Optional[float] = Field(None, alias="zwsbtol")
    research_cfexphu: Optional[float] = Field(None, alias="cfexphu")
    research_drop3d: Optional[float] = Field(None, alias="drop3d")
    research_zlayercenterbedvel: Optional[bool] = Field(
        None, alias="zlayercenterbedvel"
    )
    research_cffacver: Optional[float] = Field(None, alias="cffacver")
    research_eddyviscositybedfacmax: Optional[float] = Field(
        None, alias="eddyviscositybedfacmax"
    )
    research_epseps: Optional[float] = Field(None, alias="epseps")
    research_lateral_fixedweir_umin_method: Optional[int] = Field(
        None, alias="lateral_fixedweir_umin_method"
    )
    research_lateral_fixedweir_minimal_1d2d_embankment: Optional[float] = Field(
        None, alias="lateral_fixedweir_minimal_1d2d_embankment"
    )
    research_testfixedweirs: Optional[int] = Field(None, alias="testfixedweirs")
    research_jposhchk: Optional[int] = Field(None, alias="jposhchk")
    research_cfconhormom: Optional[float] = Field(None, alias="cfconhormom")
    research_cffachormom: Optional[float] = Field(None, alias="cffachormom")
    research_trsh_u1lb: Optional[float] = Field(None, alias="trsh_u1lb")
    research_corioconstant: Optional[int] = Field(None, alias="corioconstant")
    research_jaupwindsrc: Optional[bool] = Field(None, alias="jaupwindsrc")
    research_locsaltlev: Optional[float] = Field(None, alias="locsaltlev")
    research_subsuplupdates1: Optional[bool] = Field(None, alias="subsuplupdates1")
    research_linkdriedmx: Optional[int] = Field(None, alias="linkdriedmx")
    research_maxitpresdens: Optional[int] = Field(None, alias="maxitpresdens")
    research_lateral_fixedweir_relax: Optional[float] = Field(
        None, alias="lateral_fixedweir_relax"
    )
    research_numlimdt_baorg: Optional[int] = Field(None, alias="numlimdt_baorg")
    research_locsaltmax: Optional[float] = Field(None, alias="locsaltmax")
    research_cffachu: Optional[float] = Field(None, alias="cffachu")
    research_vertadvtypmom3onbnd: Optional[int] = Field(
        None, alias="vertadvtypmom3onbnd"
    )
    research_noderivedtypes: Optional[int] = Field(None, alias="noderivedtypes")
    research_jarhoxu: Optional[int] = Field(None, alias="jarhoxu")
    research_jaorgsethu: Optional[str] = Field(None, alias="jaorgsethu")
    research_cflwavefrac: Optional[str] = Field(None, alias="cflwavefrac")
    research_jaembed1d: Optional[str] = Field(None, alias="jaembed1d")
    research_maxitverticalforester: Optional[str] = Field(
        None, alias="maxitverticalforester"
    )
    research_ilutype: Optional[str] = Field(None, alias="ilutype")
    research_nlevel: Optional[str] = Field(None, alias="nlevel")
    research_dtol: Optional[str] = Field(None, alias="dtol")


class ResearchPhysics(Physics):
    """An extended [physics] section that includes highly experimental research keywords."""

    class Comments(Physics.Comments):
        research_surftempsmofac: Optional[str] = Field(
            "Hor. Smoothing factor for surface water in heatflx comp. (0.0-1.0), 0=no.",
            alias="surftempsmofac",
        )
        research_selfattractionloading_correct_wl_with_ini: Optional[str] = Field(
            "Correct water level with initial water level in Self attraction and loading (0=no, 1=yes).",
            alias="selfattractionloading_correct_wl_with_ini",
        )
        research_nfentrainmentmomentum: Optional[str] = Field(
            "1: Switch on momentum transfer in NearField related entrainment.",
            alias="nfentrainmentmomentum",
        )
        research_uniffrictcoef1d2d: Optional[str] = Field(
            "Uniform friction coefficient in 1D links (0: no friction).",
            alias="uniffrictcoef1d2d",
        )
        research_equili: Optional[str] = Field(
            "Equilibrium spiral flow intensity (0: no, 1: yes).", alias="equili"
        )
        research_allowcoolingbelowzero: Optional[str] = Field(
            DEPRECATED_KEYWORD, alias="allowcoolingbelowzero"
        )
        research_soiltempthick: Optional[str] = Field(
            "Use soil temperature buffer if > 0.", alias="soiltempthick"
        )
        research_selfattractionloading: Optional[str] = Field(
            "Use self attraction and loading (0: no, 1: yes, 2: only self attraction).",
            alias="selfattractionloading",
        )
        research_prandtlnumbertemperature: Optional[str] = Field(
            "Turbulent Prandtl number for temperature.",
            alias="prandtlnumbertemperature",
        )
        research_schmidtnumbersalinity: Optional[str] = Field(
            "Turbulent Schmidt number for salinity.",
            alias="schmidtnumbersalinity",
        )
        research_schmidtnumbertracer: Optional[str] = Field(
            "Turbulent Schmidt number for tracer(s).",
            alias="schmidtnumbertracer",
        )
        research_umodlin: Optional[str] = Field(
            "Linear friction umod, for ifrctyp=4,5,6.",
            alias="umodlin",
        )
        research_jadelvappos: Optional[str] = Field(
            "Only positive forced evaporation fluxes(0: no, 1: yes).",
            alias="jadelvappos",
        )
        research_uniffrictcoef1dgrlay: Optional[str] = Field(
            "Uniform ground layer friction coefficient for ocean models (m/s) (0: no friction).",
            alias="uniffrictcoef1dgrlay",
        )
        research_effectspiral: Optional[str] = Field(
            DEPRECATED_KEYWORD,
            alias="effectspiral",
        )

    comments: Comments = Comments()

    research_surftempsmofac: Optional[float] = Field(None, alias="surftempsmofac")
    research_selfattractionloading_correct_wl_with_ini: Optional[bool] = Field(
        None, alias="selfattractionloading_correct_wl_with_ini"
    )
    research_nfentrainmentmomentum: Optional[bool] = Field(
        None, alias="nfentrainmentmomentum"
    )
    research_uniffrictcoef1d2d: Optional[float] = Field(None, alias="uniffrictcoef1d2d")
    research_equili: Optional[bool] = Field(None, alias="equili")
    research_allowcoolingbelowzero: Optional[bool] = Field(
        None, alias="allowcoolingbelowzero"
    )
    research_soiltempthick: Optional[float] = Field(None, alias="soiltempthick")
    research_selfattractionloading: Optional[int] = Field(
        None, alias="selfattractionloading"
    )
    research_prandtlnumbertemperature: Optional[float] = Field(
        None, alias="prandtlnumbertemperature"
    )
    research_schmidtnumbersalinity: Optional[float] = Field(
        None, alias="schmidtnumbersalinity"
    )
    research_schmidtnumbertracer: Optional[float] = Field(
        None, alias="schmidtnumbertracer"
    )
    research_umodlin: Optional[float] = Field(None, alias="umodlin")
    research_jadelvappos: Optional[bool] = Field(None, alias="jadelvappos")
    research_uniffrictcoef1dgrlay: Optional[float] = Field(
        None, alias="uniffrictcoef1dgrlay"
    )
    research_effectspiral: Optional[float] = Field(None, alias="effectspiral")


class ResearchSediment(Sediment):
    """An extended [sediment] section that includes highly experimental research keywords."""

    class Comments(Sediment.Comments):
        research_mxgrkrone: Optional[str] = Field(
            "Highest fraction index treated by Krone.", alias="mxgrkrone"
        )
        research_seddenscoupling: Optional[str] = Field(
            "Sed rho coupling (0=no, 1=yes).", alias="seddenscoupling"
        )
        research_implicitfallvelocity: Optional[str] = Field(
            "1=Impl., 0 = Expl.", alias="implicitfallvelocity"
        )
        research_nr_of_sedfractions: Optional[str] = Field(
            "Nr of sediment fractions, (specify the next parameters for each fraction).",
            alias="nr_of_sedfractions",
        )

    comments: Comments = Comments()

    research_mxgrkrone: Optional[int] = Field(None, alias="mxgrkrone")
    research_seddenscoupling: Optional[bool] = Field(None, alias="seddenscoupling")
    research_implicitfallvelocity: Optional[int] = Field(
        None, alias="implicitfallvelocity"
    )
    research_nr_of_sedfractions: Optional[int] = Field(None, alias="nr_of_sedfractions")


class ResearchWind(Wind):
    """An extended [wind] section that includes highly experimental research keywords."""

    class Comments(Wind.Comments):
        research_windhuorzwsbased: Optional[str] = Field(
            "Wind hu or zws based, 0 = hu, 1 = zws.", alias="windhuorzwsbased"
        )
        research_varyingairdensity: Optional[str] = Field(
            "Compute air density yes/no (), 1/0, default 0.", alias="varyingairdensity"
        )
        research_wind_eachstep: Optional[str] = Field(
            "1=wind (and air pressure) each computational timestep, 0=wind (and air pressure) each usertimestep.",
            alias="wind_eachstep",
        )
        research_gapres: Optional[str] = Field(
            DEPRECATED_KEYWORD,
            alias="gapres",
        )

    comments: Comments = Comments()

    research_windhuorzwsbased: Optional[int] = Field(None, alias="windhuorzwsbased")
    research_varyingairdensity: Optional[bool] = Field(None, alias="varyingairdensity")
    research_wind_eachstep: Optional[int] = Field(None, alias="wind_eachstep")
    research_gapres: Optional[str] = Field(None, alias="gapres")


class ResearchWaves(Waves):
    """An extended [waves] section that includes highly experimental research keywords."""

    class Comments(Waves.Comments):
        research_waveswartdelwaq: Optional[str] = Field(
            "If WaveSwartDelwaq == 1 .and. Tiwaq > 0 then increase tauwave to Delwaq with 0.5rhofwuorbuorb.",
            alias="waveswartdelwaq",
        )
        research_hwavuni: Optional[str] = Field(
            "Root mean square wave height (m).", alias="hwavuni"
        )
        research_tifetchcomp: Optional[str] = Field(
            "Time interval fetch comp (s) in wavemodel 1, 2.", alias="tifetchcomp"
        )
        research_phiwavuni: Optional[str] = Field(
            "Root mean square wave direction, (deg), math convention.",
            alias="phiwavuni",
        )
        research_threedwavestreaming: Optional[str] = Field(
            "Influence of wave streaming. 0: no, 1: added to adve.",
            alias="threedwavestreaming",
        )
        research_threedwaveboundarylayer: Optional[str] = Field(
            "Boundary layer formulation. 1: Sana.", alias="threedwaveboundarylayer"
        )
        research_twavuni: Optional[str] = Field(
            "Root mean square wave period (s).", alias="twavuni"
        )
        research_uorbfac: Optional[str] = Field(
            "Orbital velocities: 0=D3D style; 1=Guza style.", alias="uorbfac"
        )
        research_threedstokesprofile: Optional[str] = Field(
            "Stokes profile. 0: no, 1:uniform over depth, 2: 2nd order Stokes theory; 3: 2, with vertical stokes gradient in adve.",
            alias="threedstokesprofile",
        )
        research_jamapsigwav: Optional[str] = Field(
            "1: sign wave height on map output; 0: hrms wave height on map output. Default=0 (legacy behaviour).",
            alias="jamapsigwav",
        )
        research_hminlw: Optional[str] = Field(
            "Cut-off depth for application of wave forces in momentum balance.",
            alias="hminlw",
        )
        research_jahissigwav: Optional[str] = Field(
            "Sign wave height on his output; 0: hrms wave height on his output. Default=1.",
            alias="jahissigwav",
        )
        research_wavenikuradse: Optional[str] = Field(
            DEPRECATED_KEYWORD,
            alias="wavenikuradse",
        )

    comments: Comments = Comments()

    research_waveswartdelwaq: Optional[bool] = Field(None, alias="waveswartdelwaq")
    research_hwavuni: Optional[float] = Field(None, alias="hwavuni")
    research_tifetchcomp: Optional[float] = Field(None, alias="tifetchcomp")
    research_phiwavuni: Optional[float] = Field(None, alias="phiwavuni")
    research_threedwavestreaming: Optional[int] = Field(
        None, alias="threedwavestreaming"
    )
    research_threedwaveboundarylayer: Optional[int] = Field(
        None, alias="threedwaveboundarylayer"
    )
    research_twavuni: Optional[float] = Field(None, alias="twavuni")
    research_uorbfac: Optional[int] = Field(None, alias="uorbfac")
    research_threedstokesprofile: Optional[int] = Field(
        None, alias="threedstokesprofile"
    )
    research_jamapsigwav: Optional[int] = Field(None, alias="jamapsigwav")
    research_hminlw: Optional[float] = Field(None, alias="hminlw")
    research_jahissigwav: Optional[int] = Field(None, alias="jahissigwav")
    research_wavenikuradse: Optional[str] = Field(None, alias="wavenikuradse")


class ResearchTime(Time):
    """An extended [time] section that includes highly experimental research keywords."""

    class Comments(Time.Comments):
        research_timestepanalysis: Optional[str] = Field(
            "Write time steps analysis file *.steps (0: no, 1: yes).",
            alias="timestepanalysis",
        )
        research_autotimestepvisc: Optional[str] = Field(
            "0 = no, 1 = yes (Time limitation based on explicit diffusive term).",
            alias="autotimestepvisc",
        )
        research_tstarttlfsmo: Optional[str] = Field(
            "Start time of smoothing of boundary conditions (Tlfsmo) w.r.t. RefDate (in TUnit).",
            alias="tstarttlfsmo",
        )

    comments: Comments = Comments()

    research_timestepanalysis: Optional[bool] = Field(None, alias="timestepanalysis")
    research_autotimestepvisc: Optional[bool] = Field(None, alias="autotimestepvisc")
    research_tstarttlfsmo: Optional[float] = Field(None, alias="tstarttlfsmo")


class ResearchRestart(Restart):
    """An extended [restart] section that includes highly experimental research keywords."""

    class Comments(Restart.Comments):
        research_rstignorebl: Optional[str] = Field(
            "Flag indicating whether bed level from restart should be ignored (0=no (default), 1=yes).",
            alias="rstignorebl",
        )

    comments: Comments = Comments()

    research_rstignorebl: Optional[bool] = Field(None, alias="rstignorebl")


class ResearchTrachytopes(Trachytopes):
    """An extended [trachytopes] section that includes highly experimental research keywords."""

    class Comments(Trachytopes.Comments):
        research_trtmth: Optional[str] = Field(
            "Area averaging method, (1=Nikuradse k based, 2=Chezy C based (parallel and serial)).",
            alias="trtmth",
        )
        research_trtmnh: Optional[str] = Field(
            "Minimum water depth for roughness computations.", alias="trtmnh"
        )
        research_trtcll: Optional[str] = Field(
            "Calibration factor file for roughness from trachytopes (see also [calibration] block).",
            alias="trtcll",
        )

    comments: Comments = Comments()

    research_trtmth: Optional[int] = Field(None, alias="trtmth")
    research_trtmnh: Optional[float] = Field(None, alias="trtmnh")
    research_trtcll: Optional[DiskOnlyFileModel] = Field(None, alias="trtcll")

    @field_validator("research_trtcll", mode="before")
    @classmethod
    def resolve_trachytopes_file_model(
        cls, value, info: ValidationInfo
    ) -> Dict[str, Any]:
        """Resolve disk-only file models."""
        return PathToDictionaryConverter.convert(cls, value, info)


class ResearchOutput(Output):
    """An extended [output] section that includes highly experimental research keywords."""

    class Comments(Output.Comments):
        research_mbalumpsourcesinks: Optional[str] = Field(
            "Lump MBA source/sink mass balance terms (1: yes, 0: no).",
            alias="mbalumpsourcesinks",
        )
        research_wrimap_nearfield: Optional[str] = Field(
            "Write near field parameters (1: yes, 0: no).", alias="wrimap_nearfield"
        )

        research_writedfminterpretedvalues: Optional[str] = Field(
            "Write DFMinterpretedvalues (1: yes, 0: no).",
            alias="writedfminterpretedvalues",
        )
        research_deleteobspointsoutsidegrid: Optional[str] = Field(
            "0 - do not delete, 1 - delete.", alias="deleteobspointsoutsidegrid"
        )
        research_mbalumpboundaries: Optional[str] = Field(
            "Lump MBA boundary mass balance terms (1: yes, 0: no).",
            alias="mbalumpboundaries",
        )
        research_waqhoraggr: Optional[str] = Field(
            "DELWAQ output horizontal aggregation file (*.dwq).", alias="waqhoraggr"
        )
        research_writedetailedtimers: Optional[str] = Field(
            "Write detailed timers output file (1: yes, 0: no).",
            alias="writedetailedtimers",
        )
        research_metadatafile: Optional[str] = Field(
            "Metadata NetCDF file with user-defined global dataset attributes (*_meta.nc).",
            alias="metadatafile",
        )
        research_mbainterval: Optional[str] = Field(
            "Mass balance area output interval (s).", alias="mbainterval"
        )
        research_wrirst_bnd: Optional[str] = Field(
            "Write waterlevel", alias="wrirst_bnd"
        )
        research_generateuuid: Optional[str] = Field(
            "Generate UUID as unique dataset identifier and include in output NetCDF files.",
            alias="generateuuid",
        )
        research_timesplitinterval: Optional[str] = Field(
            "Time splitting interval, after which a new output file is started. value+unit, e.g. '1 M', valid units: Y,M,D,h,m,s.",
            alias="timesplitinterval",
        )
        research_rugfile: Optional[str] = Field(
            "Polyline file *_rug.pli defining runup gauges.", alias="rugfile"
        )
        research_mbawritecsv: Optional[str] = Field(
            "Write mass balance area output to a csv-file (1: yes, 0: no).",
            alias="mbawritecsv",
        )
        research_mbalumpfromtomba: Optional[str] = Field(
            "Lump MBA from/to other areas mass balance terms (1: yes, 0: no).",
            alias="mbalumpfromtomba",
        )
        research_mbalumpprocesses: Optional[str] = Field(
            "Lump MBA processes mass balance terms (1: yes, 0: no).",
            alias="mbalumpprocesses",
        )
        research_waqvertaggr: Optional[str] = Field(
            "DELWAQ output vertical aggregation file (*.vag).", alias="waqvertaggr"
        )
        research_mbawritenetcdf: Optional[str] = Field(
            "Write mass balance area output to a netCDF-file (1: yes, 0: no).",
            alias="mbawritenetcdf",
        )
        research_mbawritetxt: Optional[str] = Field(
            "Write mass balance area output to a txt-file (1: yes, 0: no).",
            alias="mbawritetxt",
        )
        research_nccompression: Optional[str] = Field(
            "Whether or not (1/0) to apply compression to NetCDF output files - NOTE: only works when NcFormat = 4.",
            alias="nccompression",
        )
        research_wrimap_ice: Optional[str] = Field(
            "Write output to map file for ice cover, 0=no (default), 1=yes.",
            alias="wrimap_ice",
        )
        research_wrimap_trachytopes: Optional[str] = Field(
            "Write trachytope roughnesses to map file (1: yes, 0: no).",
            alias="wrimap_trachytopes",
        )
        research_s1incinterval: Optional[str] = Field(
            DEPRECATED_KEYWORD,
            alias="s1incinterval",
        )
        research_waqfilebase: Optional[str] = Field(
            DEPRECATED_KEYWORD,
            alias="waqfilebase",
        )
        research_snapshotdir: Optional[str] = Field(
            DEPRECATED_KEYWORD,
            alias="snapshotdir",
        )
        research_heatfluxesonoutput: Optional[str] = Field(
            DEPRECATED_KEYWORD,
            alias="heatfluxesonoutput",
        )

    comments: Comments = Comments()

    research_mbalumpsourcesinks: Optional[bool] = Field(
        None, alias="mbalumpsourcesinks"
    )
    research_wrimap_nearfield: Optional[bool] = Field(None, alias="wrimap_nearfield")
    research_writedfminterpretedvalues: Optional[bool] = Field(
        None, alias="writedfminterpretedvalues"
    )
    research_deleteobspointsoutsidegrid: Optional[bool] = Field(
        None, alias="deleteobspointsoutsidegrid"
    )
    research_mbalumpboundaries: Optional[bool] = Field(None, alias="mbalumpboundaries")
    research_waqhoraggr: Optional[DiskOnlyFileModel] = Field(None, alias="waqhoraggr")
    research_writedetailedtimers: Optional[bool] = Field(
        None, alias="writedetailedtimers"
    )
    research_metadatafile: Optional[DiskOnlyFileModel] = Field(
        None, alias="metadatafile"
    )
    research_mbainterval: Optional[float] = Field(None, alias="mbainterval")
    research_wrirst_bnd: Optional[bool] = Field(None, alias="wrirst_bnd")
    research_generateuuid: Optional[bool] = Field(None, alias="generateuuid")
    research_timesplitinterval: Optional[str] = Field(None, alias="timesplitinterval")
    research_rugfile: Optional[PolyFile] = Field(None, alias="rugfile")
    research_mbawritecsv: Optional[bool] = Field(None, alias="mbawritecsv")
    research_mbalumpfromtomba: Optional[bool] = Field(None, alias="mbalumpfromtomba")
    research_mbalumpprocesses: Optional[bool] = Field(None, alias="mbalumpprocesses")
    research_waqvertaggr: Optional[DiskOnlyFileModel] = Field(None, alias="waqvertaggr")
    research_mbawritenetcdf: Optional[bool] = Field(None, alias="mbawritenetcdf")
    research_mbawritetxt: Optional[bool] = Field(None, alias="mbawritetxt")
    research_nccompression: Optional[bool] = Field(None, alias="nccompression")
    research_wrimap_ice: Optional[bool] = Field(None, alias="wrimap_ice")
    research_wrimap_trachytopes: Optional[bool] = Field(
        None, alias="wrimap_trachytopes"
    )
    research_s1incinterval: Optional[str] = Field(None, alias="s1incinterval")
    research_waqfilebase: Optional[str] = Field(None, alias="waqfilebase")
    research_snapshotdir: Optional[str] = Field(None, alias="snapshotdir")
    research_heatfluxesonoutput: Optional[str] = Field(None, alias="heatfluxesonoutput")

    @field_validator(
        "research_waqhoraggr",
        "research_waqvertaggr",
        "research_metadatafile",
        mode="before",
    )
    @classmethod
    def _resolve_output_file_model(cls, v, info: ValidationInfo):
        return PathToDictionaryConverter.convert(cls, v, info)


class ResearchProcesses(Processes):
    """An extended [processes] section that includes highly experimental research keywords."""

    class Comments(Processes.Comments):
        research_substancedensitycoupling: Optional[str] = Field(
            "Substance rho coupling (0=no, 1=yes).", alias="substancedensitycoupling"
        )

    comments: Comments = Comments()

    research_substancedensitycoupling: Optional[bool] = Field(
        None, alias="substancedensitycoupling"
    )


class ResearchSedtrails(INIBasedModel):
    """The `[Sedtrails]` section in an MDU file."""

    class Comments(INIBasedModel.Comments):
        research_sedtrailsgrid: Optional[str] = Field(
            "Grid file for sedtrails output locations on corners.",
            alias="sedtrailsgrid",
        )
        research_sedtrailsanalysis: Optional[str] = Field(
            "Sedtrails analysis. Should be all, transport, flowvelocity or soulsby.",
            alias="sedtrailsanalysis",
        )
        research_sedtrailsinterval: Optional[str] = Field(
            "Sedtrails output times (s), interval, starttime, stoptime (s), if starttime, stoptime are left blank, use whole simulation period.",
            alias="sedtrailsinterval",
        )
        research_sedtrailsoutputfile: Optional[str] = Field(
            "Sedtrails time-avgd output file.", alias="sedtrailsoutputfile"
        )

    comments: Comments = Comments()
    _header: Literal["sedtrails"] = "sedtrails"

    research_sedtrailsgrid: Optional[DiskOnlyFileModel] = Field(
        None, alias="sedtrailsgrid"
    )
    research_sedtrailsanalysis: Optional[
        Literal["all", "transport", "flowvelocity", "soulsby"]
    ] = Field(None, alias="sedtrailsanalysis")
    research_sedtrailsinterval: Optional[List[float]] = Field(
        None, alias="sedtrailsinterval"
    )
    research_sedtrailsoutputfile: Optional[DiskOnlyFileModel] = Field(
        None, alias="sedtrailsoutputfile"
    )

    @field_validator(
        "research_sedtrailsgrid", "research_sedtrailsoutputfile", mode="before"
    )
    @classmethod
    def _resolve_sedtrails_file_model(cls, v, info: ValidationInfo):
        return PathToDictionaryConverter.convert(cls, v, info)

    @field_validator("research_sedtrailsinterval", mode="before")
    @classmethod
    def _split_to_list(cls, v, info: ValidationInfo) -> List[float]:
        return split_string_on_delimiter(cls, v, info)


class ResearchFMModel(FMModel):
    """
    An extended FMModel that includes highly experimental research sections and keywords.
    """

    general: ResearchGeneral = Field(default_factory=ResearchGeneral)
    geometry: ResearchGeometry = Field(default_factory=ResearchGeometry)
    numerics: ResearchNumerics = Field(default_factory=ResearchNumerics)
    physics: ResearchPhysics = Field(default_factory=ResearchPhysics)
    sediment: ResearchSediment = Field(default_factory=ResearchSediment)
    wind: ResearchWind = Field(default_factory=ResearchWind)
    waves: Optional[ResearchWaves] = Field(None)
    time: ResearchTime = Field(default_factory=ResearchTime)
    restart: ResearchRestart = Field(default_factory=ResearchRestart)
    trachytopes: ResearchTrachytopes = Field(default_factory=ResearchTrachytopes)
    output: ResearchOutput = Field(default_factory=ResearchOutput)
    processes: Optional[ResearchProcesses] = Field(None)
    sedtrails: Optional[ResearchSedtrails] = Field(None)



"""models.py defines the RainfallRunoffModel and supporting structures."""

from pathlib import Path
from typing import Annotated, Any, Callable, Dict, Iterable, Optional

from pydantic import (
    BeforeValidator,
    ConfigDict,
    Field,
    FilePath,
    ValidationInfo,
    field_validator,
    model_validator,
)

from hydrolib.core.base.models import (
    DiskOnlyFileModel,
    ModelSaveSettings,
    ParsableFileModel,
    SerializerConfig,
    set_default_disk_only_file_model,
)

from .meteo.models import BuiModel
from .parser import read
from .serializer import write
from .topology.models import LinkFile, NodeFile


class ImmutableDiskOnlyFileModel(DiskOnlyFileModel):
    """
    ImmutableDiskOnlyFileModel modifies the DiskOnlyFileModel to provide faux
    immutablitity.

    This behaviour is required for the mappix properties, which should always
    have the same name and path and should not be modified by users.
    """

    model_config = ConfigDict(frozen=True)


_mappix_paved_area_sewage_storage_name = "pvstordt.his"
_mappix_paved_area_flow_rates_name = "pvflowdt.his"
_mappix_unpaved_area_flow_rates_name = "upflowdt.his"
_mappix_ground_water_levels_name = "upgwlvdt.his"
_mappix_green_house_bassins_storage_name = "grnstrdt.his"
_mappix_green_house_bassins_results_name = "grnflodt.his"
_mappix_open_water_details_name = "ow_lvldt.his"
_mappix_exceedance_time_reference_levels_name = "ow_excdt.his"
_mappix_flow_rates_over_structures_name = "strflodt.his"
_mappix_flow_rates_to_edge_name = "bndflodt.his"
_mappix_pluvius_max_sewage_storage_name = "plvstrdt.his"
_mappix_pluvius_max_flow_rates_name = "plvflodt.his"
_mappix_balance_name = "balansdt.his"
_mappix_cumulative_balance_name = "cumbaldt.his"
_mappix_salt_concentrations_name = "saltdt.his"


def validate_mappix_value(field_name: str, expected: str, v) -> classmethod:
    if str(v.filepath) != expected:
        actual_value = v or "None"
        raise ValueError(
            f"Expected a value of '{expected}' for '{field_name}' but got '{actual_value}'. Mappix values should not be changed."
        )
    return v


class RainfallRunoffModel(ParsableFileModel):
    """The RainfallRunoffModel contains all paths and sub-models related to the
    Rainfall Runoff model.
    """

    # Note that order is defined by the .fnm file type and is used for parsing the data.
    control_file: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("delft_3b.ini")))
    node_data: Optional[NodeFile] = Field(None)
    link_data: Optional[LinkFile] = Field(None)
    open_water_data: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("3brunoff.tp")))
    paved_area_general: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("paved.3b")))
    paved_area_storage: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("paved.sto")))
    paved_area_dwa: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("paved.dwa")))
    paved_area_sewer_pump_capacity: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("paved.tbl")))
    boundaries: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("pluvius.dwa")))
    pluvius: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("pluvius.3b")))
    pluvius_general: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("pluvius.alg")))
    kasklasse: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("kasklass")))
    bui_file: Optional[BuiModel] = Field(None)
    verdampings_file: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("default.evp")))
    unpaved_area_general: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("unpaved.3b")))
    unpaved_area_storage: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("unpaved.sto")))
    green_house_area_initialisation: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("kasinit")))
    green_house_area_usage_data: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("kasgebr")))
    crop_factors: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("cropfact")))
    table_bergingscoef: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("bergcoef")))
    unpaved_alpha_factor_definitions: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("unpaved.alf")))
    run_log_file: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("sobek_3b.log")))
    schema_overview: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("3b_gener.out")))
    output_results_paved_area: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("paved.out")))
    output_results_unpaved_area: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("unpaved.out")))
    output_results_green_houses: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("grnhous.out")))
    output_results_open_water: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("openwate.out")))
    output_results_structures: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("struct3b.out")))
    output_results_boundaries: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("bound3b.out")))
    output_results_pluvius: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("pluvius.out")))
    infiltration_definitions_unpaved: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("unpaved.inf")))
    run_debug_file: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("sobek_3b.dbg")))
    unpaved_seepage: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("unpaved.sep")))
    unpaved_initial_values_tables: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("unpaved.tbl")))
    green_house_general: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("greenhse.3b")))
    green_house_roof_storage: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("greenhse.rf")))
    pluvius_sewage_entry: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("runoff.out")))
    input_variable_gauges_on_edge_nodes: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("sbk_rtc.his")))
    input_salt_data: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("salt.3b")))
    input_crop_factors_open_water: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("crop_ow.prn")))
    restart_file_input: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("RSRR_IN")))
    restart_file_output: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("RSRR_OUT")))
    binary_file_input: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("3b_input.bin")))
    sacramento_input: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("sacrmnto.3b")))
    output_flow_rates_edge_nodes: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("aanvoer.abr")))
    output_salt_concentration_edge: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("saltbnd.out")))
    output_salt_exportation: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("salt.out")))
    greenhouse_silo_definitions: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("greenhse.sil")))
    open_water_general: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("openwate.3b")))
    open_water_seepage_definitions: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("openwate.sep")))
    open_water_tables_target_levels: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("openwate.tbl")))
    structure_general: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("struct3b.dat")))
    structure_definitions: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("struct3b.def")))
    controller_definitions: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("contr3b.def")))
    structure_tables: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("struct3b.tbl")))
    boundary_data: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("bound3b.3b")))
    boundary_tables: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("bound3b.tbl")))
    sobek_location_rtc: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("sbk_loc.rtc")))
    wwtp_data: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("wwtp.3b")))
    wwtp_tables: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("wwtp.tbl")))
    industry_general: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("industry.3b")))

    # Mappix paths should not change and always hold the defined values.
    @field_validator("mappix_paved_area_sewage_storage", mode="before")
    @classmethod
    def _validate_mappix_paved_area_sewage_storage(
        cls, v: Any, info: ValidationInfo
    ) -> Any:
        return validate_mappix_value(
            info.field_name, _mappix_paved_area_sewage_storage_name, v
        )

    mappix_paved_area_sewage_storage: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_paved_area_sewage_storage_name)),
        frozen=True,
    )

    @field_validator("mappix_paved_area_flow_rates", mode="before")
    @classmethod
    def _validate_mappix_paved_area_flow_rates(
        cls, v: Any, info: ValidationInfo
    ) -> Any:
        return validate_mappix_value(
            info.field_name, _mappix_paved_area_flow_rates_name, v
        )

    mappix_paved_area_flow_rates: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_paved_area_flow_rates_name)),
        frozen=True,
    )

    @field_validator("mappix_unpaved_area_flow_rates", mode="before")
    @classmethod
    def _validate_mappix_unpaved_area_flow_rates(
        cls, v: Any, info: ValidationInfo
    ) -> Any:
        return validate_mappix_value(
            info.field_name, _mappix_unpaved_area_flow_rates_name, v
        )

    mappix_unpaved_area_flow_rates: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_unpaved_area_flow_rates_name)),
        frozen=True,
    )

    @field_validator("mappix_ground_water_levels", mode="before")
    @classmethod
    def _validate_mappix_ground_water_levels(cls, v: Any, info: ValidationInfo) -> Any:
        return validate_mappix_value(
            info.field_name, _mappix_ground_water_levels_name, v
        )

    mappix_ground_water_levels: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_ground_water_levels_name)),
        frozen=True,
    )

    @field_validator("mappix_green_house_bassins_storage", mode="before")
    @classmethod
    def _validate_mappix_green_house_bassins_storage(
        cls, v: Any, info: ValidationInfo
    ) -> Any:
        return validate_mappix_value(
            info.field_name, _mappix_green_house_bassins_storage_name, v
        )

    mappix_green_house_bassins_storage: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_green_house_bassins_storage_name)),
        frozen=True,
    )

    @field_validator("mappix_green_house_bassins_results", mode="before")
    @classmethod
    def _validate_mappix_green_house_bassins_results(
        cls, v: Any, info: ValidationInfo
    ) -> Any:
        return validate_mappix_value(
            info.field_name, _mappix_green_house_bassins_results_name, v
        )

    mappix_green_house_bassins_results: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_green_house_bassins_results_name)),
        frozen=True,
    )

    @field_validator("mappix_open_water_details", mode="before")
    @classmethod
    def _validate_mappix_open_water_details(cls, v: Any, info: ValidationInfo) -> Any:
        return validate_mappix_value(
            info.field_name, _mappix_open_water_details_name, v
        )

    mappix_open_water_details: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_open_water_details_name)),
        frozen=True,
    )

    @field_validator("mappix_exceedance_time_reference_levels", mode="before")
    @classmethod
    def _validate_mappix_exceedance_time_reference_levels(
        cls, v: Any, info: ValidationInfo
    ) -> Any:
        return validate_mappix_value(
            info.field_name, _mappix_exceedance_time_reference_levels_name, v
        )

    mappix_exceedance_time_reference_levels: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_exceedance_time_reference_levels_name)),
        frozen=True,
    )

    @field_validator("mappix_flow_rates_over_structures", mode="before")
    @classmethod
    def _validate_mappix_flow_rates_over_structures(
        cls, v: Any, info: ValidationInfo
    ) -> Any:
        return validate_mappix_value(
            info.field_name, _mappix_flow_rates_over_structures_name, v
        )

    mappix_flow_rates_over_structures: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_flow_rates_over_structures_name)),
        frozen=True,
    )

    @field_validator("mappix_flow_rates_to_edge", mode="before")
    @classmethod
    def _validate_mappix_flow_rates_to_edge(cls, v: Any, info: ValidationInfo) -> Any:
        return validate_mappix_value(
            info.field_name, _mappix_flow_rates_to_edge_name, v
        )

    mappix_flow_rates_to_edge: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_flow_rates_to_edge_name)),
        frozen=True,
    )

    @field_validator("mappix_pluvius_max_sewage_storage", mode="before")
    @classmethod
    def _validate_mappix_pluvius_max_sewage_storage(
        cls, v: Any, info: ValidationInfo
    ) -> Any:
        return validate_mappix_value(
            info.field_name, _mappix_pluvius_max_sewage_storage_name, v
        )

    mappix_pluvius_max_sewage_storage: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_pluvius_max_sewage_storage_name)),
        frozen=True,
    )

    @field_validator("mappix_pluvius_max_flow_rates", mode="before")
    @classmethod
    def _validate_mappix_pluvius_max_flow_rates(
        cls, v: Any, info: ValidationInfo
    ) -> Any:
        return validate_mappix_value(
            info.field_name, _mappix_pluvius_max_flow_rates_name, v
        )

    mappix_pluvius_max_flow_rates: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_pluvius_max_flow_rates_name)),
        frozen=True,
    )

    @field_validator("mappix_balance", mode="before")
    @classmethod
    def _validate_mappix_balance(cls, v: Any, info: ValidationInfo) -> Any:
        return validate_mappix_value(info.field_name, _mappix_balance_name, v)

    mappix_balance: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_balance_name)),
        frozen=True,
    )

    @field_validator("mappix_cumulative_balance", mode="before")
    @classmethod
    def _validate_mappix_cumulative_balance(cls, v: Any, info: ValidationInfo) -> Any:
        return validate_mappix_value(
            info.field_name, _mappix_cumulative_balance_name, v
        )

    mappix_cumulative_balance: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_cumulative_balance_name)),
        frozen=True,
    )

    @field_validator("mappix_salt_concentrations", mode="before")
    @classmethod
    def _validate_mappix_salt_concentrations(cls, v: Any, info: ValidationInfo) -> Any:
        return validate_mappix_value(
            info.field_name, _mappix_salt_concentrations_name, v
        )

    mappix_salt_concentrations: ImmutableDiskOnlyFileModel = Field(
        ImmutableDiskOnlyFileModel(Path(_mappix_salt_concentrations_name)),
        frozen=True,
    )

    industry_tables: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("industry.tbl")))
    maalstop: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("rtc_3b.his")))
    time_series_temperature: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("default.tmp")))
    time_series_runoff: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("rnff.#")))
    discharges_totals_at_edge_nodes: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("bndfltot.his")))
    language_file: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("sobek_3b.lng")))
    ow_volume: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("ow_vol.his")))
    ow_levels: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("ow_level.his")))
    balance_file: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("3b_bal.out")))
    his_3b_area_length: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("3bareas.his")))
    his_3b_structure_data: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("3bstrdim.his")))
    his_rr_runoff: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("rrrunoff.his")))
    his_sacramento: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("sacrmnto.his")))
    his_rwzi: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("wwtpdt.his")))
    his_industry: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("industdt.his")))
    ctrl_ini: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("ctrl.ini")))
    root_capsim_input_file: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("root_sim.inp")))
    unsa_capsim_input_file: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("unsa_sim.inp")))
    capsim_message_file: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("capsim.msg")))
    capsim_debug_file: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("capsim.dbg")))
    restart_1_hour: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("restart1.out")))
    restart_12_hours: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("restart12.out")))
    rr_ready: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("RR-ready")))
    nwrw_areas: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("NwrwArea.His")))
    link_flows: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("3blinks.his")))
    modflow_rr: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("modflow_rr.His")))
    rr_modflow: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("rr_modflow.His")))
    rr_wlm_balance: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("rr_wlmbal.His")))
    sacramento_ascii_output: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("sacrmnto.out")))
    nwrw_input_dwa_table: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("pluvius.tbl")))
    rr_balance: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("rrbalans.his")))
    green_house_classes: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("KasKlasData.dat")))
    green_house_init: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("KasInitData.dat")))
    green_house_usage: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("KasGebrData.dat")))
    crop_factor: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("CropData.dat")))
    crop_ow: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("CropOWData.dat")))
    soil_data: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("SoilData.dat")))
    dio_config_ini_file: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("dioconfig.ini")))
    bui_file_for_continuous_calculation_series: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("NWRWCONT.#")))
    nwrw_output: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("NwrwSys.His")))
    rr_routing_link_definitions: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("3b_rout.3b")))
    cell_input_file: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("3b_cel.3b")))
    cell_output_file: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("3b_cel.his")))
    rr_simulate_log_file: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("sobek3b_progress.txt")))
    rtc_coupling_wq_salt: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("wqrtc.his")))
    rr_boundary_conditions_sobek3: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(Path("BoundaryConditions.bc")))

    rr_ascii_restart_openda: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(filepath=None))
    lgsi_cachefile: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(filepath=None))
    meteo_input_file_rainfall: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(filepath=None))
    meteo_input_file_evaporation: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(filepath=None))
    meteo_input_file_temperature: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(filepath=None))

    @model_validator(mode="before")
    @classmethod
    def _validate_diskonlyfilemodel(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Ensure that all DiskOnlyFileModel fields are set to a default value if None."""
        for key, value in values.items():
            field = cls.model_fields.get(key)
            if (
                field is not None
                and field.annotation is DiskOnlyFileModel
                and isinstance(value, (str, Path))
            ):
                values[key] = DiskOnlyFileModel(Path(value))
        return values

    @model_validator(mode="before")
    @classmethod
    def _validate_immutable_diskonlyfilemodel(
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Ensure that all DiskOnlyFileModel fields are set to a default value if None."""
        for key, value in values.items():
            field = cls.model_fields.get(key)
            if (
                field is not None
                and field.annotation is ImmutableDiskOnlyFileModel
                and isinstance(value, (str, Path))
            ):
                values[key] = ImmutableDiskOnlyFileModel(Path(value))
        return values

    @field_validator("node_data", mode="before")
    @classmethod
    def _validate_node_data(cls, value: Optional[str]) -> Optional[NodeFile]:
        """Ensure that node_data is set to parsed to a model if not none."""
        if value and isinstance(value, (str, Path)):
            return NodeFile(Path(value))
        return value

    @field_validator("link_data", mode="before")
    @classmethod
    def _validate_link_data(cls, value: Optional[str]) -> Optional[LinkFile]:
        """Ensure that link_data is set to parsed to a model if not none."""
        if value and isinstance(value, (str, Path)):
            return LinkFile(Path(value))
        return value

    @field_validator("bui_file", mode="before")
    @classmethod
    def _validate_bui_file(cls, value: Optional[str]) -> Optional[BuiModel]:
        """Ensure that bui_file is set to parsed to a model if not none."""
        if value and isinstance(value, (str, Path)):
            return BuiModel(Path(value))
        return value

    @classmethod
    def property_keys(cls) -> Iterable[str]:
        # Skip first two elements corresponding with file_path and serializer_config introduced by the FileModel.
        return list(cls.model_json_schema()["properties"].keys())[2:]

    @classmethod
    def _ext(cls) -> str:
        return ".fnm"

    @classmethod
    def _filename(cls) -> str:
        return "rr"

    @classmethod
    def _get_parser(cls) -> Callable[[FilePath], Dict]:
        return lambda path: read(cls.property_keys(), path)

    @classmethod
    def _get_serializer(
        cls,
    ) -> Callable[[Path, Dict, SerializerConfig, ModelSaveSettings], None]:
        return write



from typing import Literal

from pydantic import Field

from hydrolib.core.dflowfm.research.models import ResearchFMModel, ResearchGeneral


class LegacyGeneralAsModel(ResearchGeneral):
    """
    The `[Model]` section in a legacy MDU file.

    This a small helper class to support legacy versions of .mdu files, where
    the `[General]` section with metadata did not exist yet, and was called
    `[Model]` instead.
    """

    _header: Literal["Model"] = "Model"
    mduformatversion: str = Field("1.06", alias="MDUFormatVersion")


class LegacyFMModel(ResearchFMModel):
    """
    Legacy version of an MDU file, specifically for: MDUFormatVersion <= 1.06.

    This a small wrapper class to support legacy versions of .mdu files, where
    the `[General]` section with metadata did not exist yet, and was called
    `[Model]` instead.
    """

    model: LegacyGeneralAsModel = Field(default_factory=LegacyGeneralAsModel)
